# Copyright (c) 2019-2023 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
# THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.

import pytest
import numpy

import cupy
from cupy import testing
import cupyx.signal


# https://github.com/python-acoustics/python-acoustics/blob/master/acoustics/cepstrum.py
def complex_cepstrum(x, n=None):
    """Compute the complex cepstrum of a real sequence.
    Parameters
    ----------
    x : ndarray
        Real sequence to compute complex cepstrum of.
    n : {None, int}, optional
        Length of the Fourier transform.
    Returns
    -------
    ceps : ndarray
        The complex cepstrum of the real data sequence `x` computed using the
        Fourier transform.
    ndelay : int
        The amount of samples of circular delay added to `x`.
    The complex cepstrum is given by
    .. math:: c[n] = F^{-1}\\left{\\log_{10}{\\left(F{x[n]}\\right)}\\right}
    where :math:`x_[n]` is the input signal and :math:`F` and :math:`F_{-1}
    are respectively the forward and backward Fourier transform.
    --------
    """

    def _unwrap(phase):
        samples = phase.shape[-1]
        unwrapped = numpy.unwrap(phase)
        center = (samples + 1) // 2
        if samples == 1:
            center = 0
        ndelay = numpy.array(numpy.round(unwrapped[..., center] / numpy.pi))
        unwrapped -= (numpy.pi * ndelay[..., None] * numpy.arange(samples)
                      / center)
        return unwrapped, ndelay

    spectrum = numpy.fft.fft(x, n=n)
    unwrapped_phase, ndelay = _unwrap(numpy.angle(spectrum))
    log_spectrum = numpy.log(numpy.abs(spectrum)) + 1j * unwrapped_phase
    ceps = numpy.fft.ifft(log_spectrum).real

    return ceps, ndelay


def real_cepstrum(x, n=None):
    """
    Compute the real cepstrum of a real sequence.
    x : ndarray
        Real sequence to compute real cepstrum of.
    n : {None, int}, optional
        Length of the Fourier transform.
    Returns
    -------
    ceps: ndarray
        The real cepstrum.
    """
    spectrum = numpy.fft.fft(x, n=n)
    ceps = numpy.fft.ifft(numpy.log(numpy.abs(spectrum))).real

    return ceps


def inverse_complex_cepstrum(ceps, ndelay):
    r"""Compute the inverse complex cepstrum of a real sequence.
    ceps : ndarray
        Real sequence to compute inverse complex cepstrum of.
    ndelay: int
        The amount of samples of circular delay added to `x`.
    Returns
    -------
    x : ndarray
        The inverse complex cepstrum of the real sequence `ceps`.
    The inverse complex cepstrum is given by
    .. math:: x[n] = F^{-1}\left{\exp(F(c[n]))\right}
    where :math:`c_[n]` is the input signal and :math:`F` and :math:`F_{-1}
    are respectively the forward and backward Fourier transform.
    """

    def _wrap(phase, ndelay):
        ndelay = numpy.array(ndelay)
        samples = phase.shape[-1]
        center = (samples + 1) // 2
        wrapped = (
            phase + numpy.pi * ndelay[..., None] * numpy.arange(samples)
            / center
        )
        return wrapped

    log_spectrum = numpy.fft.fft(ceps)
    spectrum = numpy.exp(
        log_spectrum.real + 1j * _wrap(log_spectrum.imag, ndelay)
    )
    x = numpy.fft.ifft(spectrum).real

    return x


def minimum_phase(x, n=None):
    r"""Compute the minimum phase reconstruction of a real sequence.
    x : ndarray
        Real sequence to compute the minimum phase reconstruction of.
    n : {None, int}, optional
        Length of the Fourier transform.
    Compute the minimum phase reconstruction of a real sequence using the
    real cepstrum.
    Returns
    -------
    m : ndarray
        The minimum phase reconstruction of the real sequence `x`.
    """
    if n is None:
        n = len(x)
    ceps = real_cepstrum(x, n=n)
    odd = n % 2
    window = numpy.concatenate(
        (
            [1.0],
            2.0 * numpy.ones((n + odd) // 2 - 1),
            numpy.ones(1 - odd),
            numpy.zeros((n + odd) // 2 - 1),
        )
    )

    m = numpy.fft.ifft(numpy.exp(numpy.fft.fft(window * ceps))).real

    return m


@pytest.mark.parametrize("num_samps", [2**8, 2**14])
@pytest.mark.parametrize("n", [123, 256])
def test_complex_cepstrum(num_samps, n):
    cpu_sig = numpy.random.rand(num_samps)
    gpu_sig = cupy.array(cpu_sig)
    gpu_out = cupyx.signal.complex_cepstrum(gpu_sig, n)
    cpu_out = complex_cepstrum(cpu_sig, n)
    testing.assert_allclose(cpu_out[0], gpu_out[0])
    testing.assert_allclose(cpu_out[1], gpu_out[1])


@pytest.mark.parametrize("num_samps", [2**8, 2**14])
@pytest.mark.parametrize("n", [123, 256])
def test_real_cepstrum(num_samps, n):
    cpu_sig = numpy.random.rand(num_samps)
    gpu_sig = cupy.array(cpu_sig)
    gpu_out = cupyx.signal.real_cepstrum(gpu_sig, n)
    cpu_out = real_cepstrum(cpu_sig, n)
    testing.assert_allclose(cpu_out, gpu_out)


@pytest.mark.parametrize("num_samps", [2**10])
@pytest.mark.parametrize("n", [123, 256])
def test_inverse_complex_cepstrum(num_samps, n):
    cpu_sig = numpy.random.rand(num_samps)
    gpu_sig = cupy.array(cpu_sig)
    gpu_out = cupyx.signal.inverse_complex_cepstrum(gpu_sig, n)
    cpu_out = inverse_complex_cepstrum(cpu_sig, n)
    testing.assert_allclose(cpu_out, gpu_out)


@pytest.mark.parametrize("num_samps", [2**8, 2**14])
@pytest.mark.parametrize("n", [123, 256])
def test_minimum_phase(num_samps, n):
    cpu_sig = numpy.random.rand(num_samps)
    gpu_sig = cupy.array(cpu_sig)
    gpu_out = cupyx.signal.minimum_phase(gpu_sig, n)
    cpu_out = minimum_phase(cpu_sig, n)
    testing.assert_allclose(cpu_out, gpu_out)



import unittest

import numpy
import pytest

import cupy
from cupy import testing


@testing.parameterize(*testing.product(
    {'shape': [(3,), (2, 3, 4), (0,), (0, 2), (3, 0)]},
))
class TestIter(unittest.TestCase):

    @testing.for_all_dtypes()
    @testing.numpy_cupy_array_equal()
    def test_list(self, xp, dtype):
        x = testing.shaped_arange(self.shape, xp, dtype)
        return list(x)

    @testing.for_all_dtypes()
    @testing.numpy_cupy_equal()
    def test_len(self, xp, dtype):
        x = testing.shaped_arange(self.shape, xp, dtype)
        return len(x)


class TestIterInvalid(unittest.TestCase):

    @testing.for_all_dtypes()
    def test_iter(self, dtype):
        for xp in (numpy, cupy):
            x = testing.shaped_arange((), xp, dtype)
            with pytest.raises(TypeError):
                iter(x)

    @testing.for_all_dtypes()
    def test_len(self, dtype):
        for xp in (numpy, cupy):
            x = testing.shaped_arange((), xp, dtype)
            with pytest.raises(TypeError):
                len(x)



import cupy
from cupy.linalg import _util

# Find the "bandwise position" of a nonzero cell
_kernel_cupy_band_pos_c = cupy.ElementwiseKernel(
    'T A, N r, N c',
    'N out',
    'out = A != 0 ? r - c : 0',
    'cupyx_scipy_linalg_band_pos'
)


def bandwidth(a):
    """Return the lower and upper bandwidth of a 2D numeric array.
    Parameters
    ----------
    a : ndarray
        Input array of size (M, N)
    Returns
    -------
    lu : tuple
        2-tuple of ints indicating the lower and upper bandwidth. A zero
        denotes no sub- or super-diagonal on that side (triangular), and,
        say for M rows (M-1) means that side is full. Same example applies
        to the upper triangular part with (N-1).

    .. seealso:: :func:`scipy.linalg.bandwidth`
    """

    a = cupy.asarray(a)

    if a.size == 0:
        return (0, 0)
    _util._assert_2d(a)

    # Create new matrix A which is C contiguous
    if a.flags['F_CONTIGUOUS']:
        A = a.T
    else:
        A = a

    # row_num and col_num contain info on the row and column number of A
    m, n = A.shape
    row_num, col_num = cupy.mgrid[0:m, 0:n]
    bandpts = _kernel_cupy_band_pos_c(A, row_num, col_num)

    # If F contiguous, transpose
    if a.flags['F_CONTIGUOUS']:
        upper_band = int(cupy.amax(bandpts))
        lower_band = -int(cupy.amin(bandpts))
    else:
        lower_band = int(cupy.amax(bandpts))
        upper_band = -int(cupy.amin(bandpts))

    return lower_band, upper_band



import pytest

import cupy
from cupy.cuda import runtime
from cupyx import jit


@pytest.mark.skipif(runtime.is_hip, reason="not supported on HIP")
class TestCooperativeGroups:

    def test_thread_block_group(self):
        @jit.rawkernel()
        def test_thread_block(x):
            g = jit.cg.this_thread_block()
            if g.thread_rank() == 0:
                x[0] += 101
            if g.thread_rank() == 1:
                x[1] = g.size()
            # test dim3
            if g.thread_rank() == 2:
                g_idx = g.group_index()
                x[2], x[3], x[4] = g_idx.x, g_idx.y, g_idx.z
            if g.thread_rank() == 3:
                t_idx = g.thread_index()
                x[5], x[6], x[7] = t_idx.x, t_idx.y, t_idx.z
            if g.thread_rank() == 4:
                g_dim = g.group_dim()
                x[8], x[9], x[10] = g_dim.x, g_dim.y, g_dim.z
            g.sync()

        x = cupy.empty((16,), dtype=cupy.int64)
        x[:] = -1
        test_thread_block[1, 32](x)
        assert x[0] == 100
        assert x[1] == 32
        assert (x[2], x[3], x[4]) == (0, 0, 0)
        assert (x[5], x[6], x[7]) == (3, 0, 0)
        assert (x[8], x[9], x[10]) == (32, 1, 1)
        assert (x[11:] == -1).all()

    @pytest.mark.skipif(
        runtime._getLocalRuntimeVersion() < 11060,
        reason='not supported until CUDA 11.6')
    def test_thread_block_group_cu116_new_APIs(self):
        @jit.rawkernel()
        def test_thread_block(x):
            g = jit.cg.this_thread_block()
            if g.thread_rank() == 0:
                x[0] = g.num_threads()
            if g.thread_rank() == 1:
                d_th = g.dim_threads()
                x[1], x[2], x[3] = d_th.x, d_th.y, d_th.z
            g.sync()

        x = cupy.empty((16,), dtype=cupy.int64)
        x[:] = -1
        test_thread_block[1, 32](x)
        assert x[0] == 32
        assert (x[1], x[2], x[3]) == (32, 1, 1)
        assert (x[4:] == -1).all()

    @pytest.mark.skipif(runtime.deviceGetAttribute(
        runtime.cudaDevAttrCooperativeLaunch, 0) == 0,
        reason='cooperative launch is not supported on device 0')
    def test_grid_group(self):
        @jit.rawkernel()
        def test_grid(x):
            g = jit.cg.this_grid()
            if g.thread_rank() == 0:
                x[0] = g.is_valid()
            if g.thread_rank() == 1:
                x[1] = g.size()
            if g.thread_rank() == 32:  # on the 2nd group
                # Note: this is not yet possible...
                # x[2], x[3], x[4] == g.group_dim()
                g_dim = g.group_dim()
                x[2], x[3], x[4] = g_dim.x, g_dim.y, g_dim.z
            g.sync()  # this should just work!

        x = cupy.empty((16,), dtype=cupy.uint64)
        x[:] = -1  # = 2**64-1
        test_grid[2, 32](x)
        assert x[0] == 1
        assert x[1] == 64
        assert (x[2], x[3], x[4]) == (2, 1, 1)
        assert (x[5:] == 2**64-1).all()

    @pytest.mark.skipif(
        runtime._getLocalRuntimeVersion() < 11060,
        reason='not supported until CUDA 11.6')
    @pytest.mark.skipif(runtime.deviceGetAttribute(
        runtime.cudaDevAttrCooperativeLaunch, 0) == 0,
        reason='cooperative launch is not supported on device 0')
    def test_grid_group_cu116_new_APIs(self):
        @jit.rawkernel()
        def test_grid(x):
            g = jit.cg.this_grid()
            if g.thread_rank() == 1:
                x[1] = g.num_threads()
            if g.thread_rank() == 32:
                g_dim = g.dim_blocks()
                x[2], x[3], x[4] = g_dim.x, g_dim.y, g_dim.z
            if g.thread_rank() == 33:  # on the 2nd block
                x[5] = g.block_rank()
            if g.thread_rank() == 2:
                x[6] = g.num_blocks()
            if g.thread_rank() == 34:  # on the 2nd block
                b_idx = g.block_index()
                x[7], x[8], x[9] = b_idx.x, b_idx.y, b_idx.z
            g.sync()  # this should just work!

        x = cupy.empty((16,), dtype=cupy.uint64)
        x[:] = -1  # = 2**64-1
        test_grid[2, 32](x)
        assert x[1] == 64
        assert (x[2], x[3], x[4]) == (2, 1, 1)
        assert x[5] == 1
        assert x[6] == 2
        assert (x[7], x[8], x[9]) == (1, 0, 0)
        assert (x[10:] == 2**64-1).all()

    @pytest.mark.skipif(runtime.deviceGetAttribute(
        runtime.cudaDevAttrCooperativeLaunch, 0) == 0,
        reason='cooperative launch is not supported on device 0')
    def test_cg_sync(self):
        @jit.rawkernel()
        def test_sync():
            b = jit.cg.this_thread_block()
            g = jit.cg.this_grid()
            jit.cg.sync(b)
            jit.cg.sync(g)

        test_sync[2, 64]()

    @pytest.mark.parametrize(
        'test_aligned', (True, False),
    )
    def test_cg_memcpy_async_wait_for_wait(self, test_aligned):
        @jit.rawkernel()
        def test_copy(x, y):
            # do two batches of copies to test relevant APIs
            if test_aligned:
                smem = jit.shared_memory(cupy.int32, 32*2, alignment=16)
            else:
                smem = jit.shared_memory(cupy.int32, 32*2)
            g = jit.cg.this_thread_block()
            tid = g.thread_rank()
            # int32 is 4 bytes
            if test_aligned:
                # CuPy ensures x is 256B-aligned
                jit.cg.memcpy_async(
                    g, smem, 0, x, 0, 4*32, aligned_size=16)
                jit.cg.memcpy_async(
                    g, smem, 32, x, 32, 4*32, aligned_size=16)
            else:
                jit.cg.memcpy_async(
                    g, smem, 0, x, 0, 4*32)
                jit.cg.memcpy_async(
                    g, smem, 32, x, 32, 4*32)
            jit.cg.wait_prior(g, 1)
            if tid < 32:
                y[tid] = smem[tid]
            jit.cg.wait(g)
            if 32 <= tid and tid < 64:  # can't do "32 <= tid < 64" yet...
                y[tid] = smem[tid]

        x = cupy.arange(64, dtype=cupy.int32)
        y = cupy.zeros(64, dtype=cupy.int32)
        test_copy[2, 64](x, y)
        assert (x == y).all()



#
# Code in this file is adapted from SciPy version 1.11. The code in SciPy
# contains the following notice:
#
# Code adapted from "upfirdn" python library with permission:
#
# Copyright (c) 2009, Motorola, Inc
#
# All Rights Reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
# * Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright
# notice, this list of conditions and the following disclaimer in the
# documentation and/or other materials provided with the distribution.
#
# * Neither the name of Motorola nor the names of its contributors may be
# used to endorse or promote products derived from this software without
# specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
# IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
# THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
# PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
# PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from itertools import product

import pytest

import numpy as np

from cupy import testing
from cupyx.scipy.signal._upfirdn import _output_len


def upfirdn_naive(x, h, up=1, down=1):
    """Naive upfirdn processing in Python.

    Note: arg order (x, h) differs to facilitate apply_along_axis use.
    """
    h = np.asarray(h)
    out = np.zeros(len(x) * up, x.dtype)
    out[::up] = x
    out = np.convolve(h, out)[::down][:_output_len(len(h), len(x), up, down)]
    return out


def make_case(up, down, h, x_dtype, case):
    # replacement for the UpFIRDnCase class from the SciPy tests
    rng = np.random.RandomState(17)
    h = np.atleast_1d(h)
    x = {'tiny': np.ones(1, dtype=x_dtype),
         'ones': np.ones(10, dtype=x_dtype),
         'randn': rng.randn(10).astype(x_dtype),
         'ramp': np.arange(10).astype(x_dtype),
         # XXX: add 2D / 3D cases from UpFIRDnCase
         }[case]

    if 'case' == 'randn' and x_dtype in (np.complex64, np.complex128):
        x += 1j * rng.randn(10)
    return x, h


def make_case_2D(up, down, h, x_dtype, case):
    # replacement for the UpFIRDnCase class from the SciPy tests
    rng = np.random.RandomState(17)
    h = np.atleast_1d(h)

    if case == '2D':
        # 2D, random
        size = (3, 5)
        x = rng.randn(*size).astype(x_dtype)
        if x_dtype in (np.complex64, np.complex128):
            x += 1j * rng.randn(*size)
        return x, h
    elif case == '2D_noncontig':
        # 2D, random, non-contiguous
        size = (3, 7)
        x = rng.randn(*size).astype(x_dtype)
        if x_dtype in (np.complex64, np.complex128):
            x += 1j * rng.randn(*size)
        x = x[::2, 1::3].T
        return x, h
    else:
        raise ValueError(f"unknown 2D_case, {case}.")


_UPFIRDN_TYPES = (int, np.float32, np.complex64, float, complex)

_upfirdn_modes = [
    'constant', 'wrap', 'edge', 'smooth', 'symmetric', 'reflect',
    'antisymmetric', 'antireflect', 'line',
]
_upfirdn_unsupported_mode = pytest.mark.xfail(
    reason="upfirdn `mode=...` not implemented"
)


@testing.with_requires('scipy')
class TestUpfirdn:

    @testing.numpy_cupy_allclose(scipy_name='scp')
    @pytest.mark.parametrize('len_h', [1, 2, 3, 4, 5])
    @pytest.mark.parametrize('len_x', [1, 2, 3, 4, 5])
    def test_singleton(self, xp, scp, len_h, len_x):
        # gh-9844: lengths producing expected outputs
        h = xp.zeros(len_h)
        h[len_h // 2] = 1.  # make h a delta
        x = xp.ones(len_x)
        y = scp.signal.upfirdn(h, x, 1, 1)
        return y

    @testing.numpy_cupy_allclose(scipy_name='scp')
    def test_shift_x(self, xp, scp):
        # gh-9844: shifted x can change values?
        y = scp.signal.upfirdn(xp.asarray([1, 1]), xp.asarray([1.]), 1, 1)
        y1 = scp.signal.upfirdn(xp.asarray([1, 1]), xp.asarray([0., 1.]), 1, 1)
        return y, y1

    # A bunch of lengths/factors chosen because they exposed differences
    # between the "old way" and new way of computing length, and then
    # got `expected` from MATLAB
    @testing.numpy_cupy_allclose(scipy_name='scp')
    @pytest.mark.parametrize('len_h, len_x, up, down, expected', [
        (2, 2, 5, 2, [1, 0, 0, 0]),
        (2, 3, 6, 3, [1, 0, 1, 0, 1]),
        (2, 4, 4, 3, [1, 0, 0, 0, 1]),
        (3, 2, 6, 2, [1, 0, 0, 1, 0]),
        (4, 11, 3, 5, [1, 0, 0, 1, 0, 0, 1]),
    ])
    def test_length_factors(self, xp, scp, len_h, len_x, up, down, expected):
        # gh-9844: weird factors
        h = xp.zeros(len_h)
        h[0] = 1.
        x = xp.ones(len_x)
        y = scp.signal.upfirdn(h, x, up, down)
        return y

    @testing.numpy_cupy_allclose(scipy_name='scp', atol=1e-15)
    @pytest.mark.parametrize('down, want_len', [  # lengths from MATLAB
        (2, 5015),
        (11, 912),
        (79, 127),
    ])
    @pytest.mark.parametrize('dtype', _UPFIRDN_TYPES)
    def test_vs_convolve(self, xp, scp, dtype, down, want_len):
        random_state = np.random.RandomState(17)
        size = 10000

        x = random_state.randn(size).astype(dtype)
        if dtype in (np.complex64, np.complex128):
            x += 1j * random_state.randn(size)

        x = xp.asarray(x)

        h = scp.signal.firwin(31, 1. / down, window='hamming')
        y = scp.signal.upfirdn(h, x, up=1, down=down)
        return y

    @pytest.mark.parametrize('size', [8])
    # include cases with h_len > 2*size
    @pytest.mark.parametrize('h_len', [4, 5, 26])
    @pytest.mark.parametrize(
        'mode',
        [
            pytest.param(mode, marks=_upfirdn_unsupported_mode)
            if mode != 'constant'
            else mode
            for mode in _upfirdn_modes
        ]
    )
    @pytest.mark.parametrize(
        'dtype', [np.float32, np.float64, np.complex64, np.complex128]
    )
    @testing.numpy_cupy_allclose(scipy_name='scp', rtol=1e-5)
    def test_modes(self, xp, scp, size, h_len, mode, dtype):
        random_state = np.random.RandomState(5)
        x = random_state.randn(size).astype(dtype)
        if dtype in (np.complex64, np.complex128):
            x += 1j * random_state.randn(size)
        x = xp.asarray(x)

        h = xp.arange(1, 1 + h_len, dtype=x.real.dtype)

        y = scp.signal.upfirdn(h, x, up=1, down=1, mode=mode)
        return y

    @pytest.mark.parametrize('x_dtype', _UPFIRDN_TYPES)
    @pytest.mark.parametrize('h', (1., 1j))
    @pytest.mark.parametrize('up, down', [(1, 1), (2, 2), (3, 2), (2, 3)])
    @pytest.mark.parametrize('case', ['tiny', 'ones', 'randn', 'ramp'])
    @testing.numpy_cupy_allclose(scipy_name='scp')
    def test_vs_naive_delta(self, x_dtype, h, up, down, case, xp, scp):
        x, h = make_case(up, down, h, x_dtype, case)
        x = xp.asarray(x)
        h = xp.asarray(h)
        y = scp.signal.upfirdn(h, x, up, down)
        return y

    @pytest.mark.parametrize('x_dtype', _UPFIRDN_TYPES)
    @pytest.mark.parametrize('h', (1., 1j))
    @pytest.mark.parametrize('up, down', [(1, 1), (2, 2), (3, 2), (2, 3)])
    @pytest.mark.parametrize('case', ['2D', '2D_noncontig'])
    @pytest.mark.parametrize('axis', [0, 1, -1])
    @testing.numpy_cupy_allclose(scipy_name='scp')
    def test_vs_naive_delta_2D(self, axis, x_dtype, h, up, down,
                               case, xp, scp):
        x, h = make_case_2D(up, down, h, x_dtype, case)
        x = xp.asarray(x)
        h = xp.asarray(h)
        y = scp.signal.upfirdn(h, x, up, down, axis=axis)
        return y

    @pytest.mark.parametrize('x_dtype', _UPFIRDN_TYPES)
    @pytest.mark.parametrize('h_dtype', _UPFIRDN_TYPES)
    @pytest.mark.parametrize('p_max, q_max',
                             list(product((10, 100), (10, 100))))
    @pytest.mark.parametrize('case', ['tiny', 'ones', 'randn', 'ramp'])
    @testing.numpy_cupy_allclose(scipy_name='scp', atol=1e-7)
    def test_vs_naive(self, xp, scp, case, x_dtype, h_dtype, p_max, q_max):
        n_reps = 3
        longest_h = 25
        random_state = np.random.RandomState(17)
        cases = []
        for _ in range(n_reps):
            # Randomize the up/down factors somewhat
            p_add = q_max if p_max > q_max else 1
            q_add = p_max if q_max > p_max else 1
            p = random_state.randint(p_max) + p_add
            q = random_state.randint(q_max) + q_add

            # Generate random FIR coefficients
            len_h = random_state.randint(longest_h) + 1
            h = np.atleast_1d(random_state.randint(len_h))
            h = h.astype(h_dtype)
            if h_dtype == complex:  # noqa: E721
                h += 1j * random_state.randint(len_h)
            x, h = make_case(p, q, h, x_dtype, case)
            x = xp.asarray(x)
            h = xp.asarray(x)
            y = scp.signal.upfirdn(h, x, p, q)
            cases.append(y)
        return cases


def test_output_len_long_input():
    # Regression test for scipy/gh-17375.  On Windows, a large enough input
    # that should have been well within the capabilities of 64 bit integers
    # would result in a 32 bit overflow because of a bug in Cython 0.29.32.
    len_h = 1001
    in_len = 10**8
    up = 320
    down = 441
    out_len = _output_len(len_h, in_len, up, down)
    # The expected value was computed "by hand" from the formula
    #   (((in_len - 1) * up + len_h) - 1) // down + 1
    assert out_len == 72562360



from libc.stdint cimport intptr_t


###############################################################################
# Types
###############################################################################

IF CUPY_USE_CUDA_PYTHON:
    from cuda.cnvrtc cimport *
    # Aliases for compatibillity with existing CuPy codebase.
    # TODO(kmaehashi): Remove these aliases.
    ctypedef nvrtcProgram Program

cpdef check_status(int status)

cpdef tuple getVersion()
cpdef tuple getSupportedArchs()


###############################################################################
# Program
###############################################################################

cpdef intptr_t createProgram(unicode src, unicode name, headers,
                             include_names) except? 0
cpdef destroyProgram(intptr_t prog)
cpdef compileProgram(intptr_t prog, options)
cpdef bytes getPTX(intptr_t prog)
cpdef bytes getCUBIN(intptr_t prog)
cpdef bytes getNVVM(intptr_t prog)
cpdef unicode getProgramLog(intptr_t prog)
cpdef addNameExpression(intptr_t prog, str name)
cpdef str getLoweredName(intptr_t prog, str name)



import os
import sys

from cupy_builder import Context


def preflight_check(ctx: Context) -> bool:
    if sys.platform not in ('linux', 'win32'):
        print('Error: macOS is no longer supported', file=sys.stderr)
        return False

    source_root = ctx.source_root
    is_git = os.path.isdir(os.path.join(source_root, '.git'))
    for submodule in ('third_party/cccl',
                      'third_party/jitify',
                      'third_party/dlpack'):
        dirpath = os.path.join(source_root, submodule)
        if os.path.isdir(dirpath):
            if 0 < len(os.listdir(dirpath)):
                continue
        else:
            if not is_git:
                # sdist does not contain third_party directory
                continue

        if is_git:
            msg = f'''
===========================================================================
The directory {submodule} is a git submodule but is currently empty.
Please use the command:

    $ git submodule update --init

to populate the directory before building from source.
===========================================================================
        '''
        else:
            msg = f'''
===========================================================================
The directory {submodule} is a git submodule but is currently empty.
Instead of using ZIP/TAR archive downloaded from GitHub, use

    $ git clone --recursive https://github.com/cupy/cupy.git

to get a buildable CuPy source tree.
===========================================================================
        '''

        print(msg, file=sys.stderr)
        return False
    return True



from cupy import _core


# Note: complex-valued isnan, log and log1p are all defined in
#       cupy/_core/include/cupy/complex.cuh
xlogy_definition = """

template <typename T>
static __device__ T xlogy(T x, T y) {
    if ((x == (T)0.0) && !isnan(y)) {
        return (T)0.0;
    } else {
        return x * log(y);
    }
}

"""


# Note: SciPy only defines dd->d and DD->D
xlogy = _core.create_ufunc(
    'cupy_xlogy',
    ('ee->f', 'ff->f', 'dd->d', 'FF->F', 'DD->D'),
    'out0 = out0_type(xlogy(in0, in1));',
    preamble=xlogy_definition,
    doc='''Compute ``x*log(y)`` so that the result is 0 if ``x = 0``.

    Args:
        x (cupy.ndarray): input data

    Returns:
        cupy.ndarray: values of ``x * log(y)``

    .. seealso:: :data:`scipy.special.xlogy`

    ''')


xlog1py_definition = """

template <typename T>
static __device__ T xlog1py(T x, T y) {
    if ((x == (T)0.0) && ~isnan(y)) {
        return (T)0.0;
    } else {
        return x * log1p(y);
    }
}

"""

# Note: SciPy only defines dd->d and DD->D
xlog1py = _core.create_ufunc(
    'cupy_xlog1py',
    ('ee->f', 'ff->f', 'dd->d', 'FF->F', 'DD->D'),
    'out0 = out0_type(xlog1py(in0, in1));',
    preamble=xlog1py_definition,
    doc='''Compute ``x*log1p(y)`` so that the result is 0 if ``x = 0``.

    Args:
        x (cupy.ndarray): input data

    Returns:
        cupy.ndarray: values of ``x * log1p(y)``

    .. seealso:: :data:`scipy.special.xlog1py`

    ''')



import unittest

from cupy import testing
import cupyx.scipy.special  # NOQA
import pytest

try:
    import scipy.special  # NOQA
except ImportError:
    pass


@testing.with_requires('scipy<1.16')
class TestZetac(unittest.TestCase):

    @testing.for_all_dtypes(no_complex=True)
    @testing.numpy_cupy_allclose(atol=1e-5, scipy_name='scp')
    def test_arange(self, xp, scp, dtype):

        a = testing.shaped_arange((2, 3), xp, dtype)

        return scp.special.zetac(a)

    @testing.for_all_dtypes(no_complex=True, no_bool=True)
    @testing.numpy_cupy_allclose(atol=1e-5, rtol=1e-6, scipy_name='scp')
    def test_linspace(self, xp, scp, dtype):

        if xp.dtype(dtype).kind == 'u':
            pytest.skip()
        a = xp.linspace(-30, 30, 1000, dtype=dtype)

        return scp.special.zetac(a)

    @testing.for_all_dtypes(no_complex=True, no_bool=True)
    @testing.numpy_cupy_allclose(atol=1e-5, rtol=1e-6, scipy_name='scp')
    def test_linspace_small_negative_x(self, xp, scp, dtype):

        if xp.dtype(dtype).kind == 'u':
            pytest.skip()
        a = xp.linspace(-0.01, 0, 1000, dtype=dtype)

        return scp.special.zetac(a)

    @testing.for_all_dtypes(no_complex=True)
    @testing.numpy_cupy_allclose(atol=1e-2, rtol=1e-3, scipy_name='scp')
    def test_scalar(self, xp, scp, dtype):

        return scp.special.zetac(dtype(3.5))

    @testing.for_all_dtypes(no_complex=True)
    @testing.numpy_cupy_allclose(atol=1e-2, rtol=1e-3, scipy_name='scp')
    def test_inf_and_nan(self, xp, scp, dtype):
        if xp.dtype(dtype).kind in 'iu':
            pytest.skip()
        x = xp.array([-xp.inf, xp.nan, xp.inf]).astype(dtype)
        a = xp.tile(x, (3, 3))

        return scp.special.zetac(a)



import typing
from typing import Any, Optional


def _extgcd(a: int, b: int) -> tuple[int, int]:
    # Return (g, x) with g = gcd(a, b), ax + by = g - ax.

    # c - ax - by = 0  ...  (1)
    # d - au - bv = 0  ...  (2)
    c, d = a, b
    x, u = 1, 0
    # y, v = 0, 1

    # Apply Euclid's algorithm to (c, d)
    while d:
        r = c // d
        # (1), (2) = (2), (1) - (2) * r
        c, d = d, c - d * r
        x, u = u, x - u * r
        # y, v = v, y - u * r

    return c, x


def _crt(a1: int, n1: int, a2: int, n2: int) -> Optional[tuple[int, int]]:
    # Return x, n with x == a1 (mod n1), x == a2 (mod n2), n == lcm(n1, n2).
    # Choose the minimum value for x with x >= max(a1, a2). Return None if no
    # solution exists.
    # Reference: https://en.wikipedia.org/wiki/Chinese_remainder_theorem#Existence_(constructive_proof)  # NOQA

    # m1 * n1 + m2 * n2 == g
    g, m1 = _extgcd(n1, n2)

    # x == a1 == a2 (mod g)
    if (a2 - a1) % g != 0:
        return None

    n = n1 * (n2 // g)

    # x = a1 + (a2 - a1) // g * m1 * n1 % n
    x = a1 + (a2 - a1) // g * m1 % (n // n1) * n1

    if x < a2:
        x += ((a2 - x - 1) // n + 1) * n

    return x, n


def _slice_intersection(a: slice, b: slice, length: int) -> Optional[slice]:
    # Return the intersection of slice a, b. None if they are disjoint.
    a_start, a_stop, a_step = a.indices(length)
    b_start, b_stop, b_step = b.indices(length)

    crt_result = _crt(a_start, a_step, b_start, b_step)
    if crt_result is None:
        return None
    c_start, c_step = crt_result

    c_stop = min(a_stop, b_stop)
    if c_start >= c_stop:
        return None

    return slice(c_start, c_stop, c_step)


def _index_for_subslice(a: slice, sub: slice, length: int) -> slice:
    # Return slice c such that array[a][c] == array[sub].
    # sub should be contained in a.
    a_start, a_stop, a_step = a.indices(length)
    sub_start, sub_stop, sub_step = sub.indices(length)

    c_start = (sub_start - a_start) // a_step
    # a_start + a_step * (c_stop - 1) < sub_stop
    c_stop = (sub_stop - a_start - 1) // a_step + 1
    c_step = sub_step // a_step

    return slice(c_start, c_stop, c_step)


def _index_intersection(
    a_idx: tuple[slice, ...], b_idx: tuple[slice, ...],
    shape: tuple[int, ...],
) -> Optional[tuple[slice, ...]]:
    # Return None if a, b are disjoint.
    assert len(a_idx) == len(b_idx)
    result = tuple(_slice_intersection(a, b, length)
                   for a, b, length in zip(a_idx, b_idx, shape))

    if None in result:
        return None
    else:
        return typing.cast(tuple[slice, ...], result)


def _index_for_subindex(
    a_idx: tuple[slice, ...], sub_idx: tuple[slice, ...],
    shape: tuple[int, ...],
) -> tuple[slice, ...]:
    assert len(a_idx) == len(sub_idx)

    return tuple(_index_for_subslice(a, sub, length)
                 for a, sub, length in zip(a_idx, sub_idx, shape))


def _shape_after_indexing(
    outer_shape: tuple[int, ...],
    idx: tuple[slice, ...],
) -> tuple[int, ...]:
    shape = list(outer_shape)
    for i in range(len(idx)):
        start, stop, step = idx[i].indices(shape[i])
        shape[i] = (stop - start - 1) // step + 1
    return tuple(shape)


def _normalize_index(shape: tuple[int, ...], idx: Any) -> tuple[slice, ...]:
    # Convert idx to type tuple[slice, ...] with length == ndim.
    # start, stop, step of each slice are set to a non-None value.
    if not isinstance(idx, tuple):
        idx = (idx,)

    ndim = len(shape)
    if len(idx) > ndim:
        raise IndexError(
            'too many indices for array:'
            f' array is {ndim}-dimensional, but {len(idx)} were indexed')
    idx = idx + (slice(None),) * (ndim - len(idx))

    new_idx = []
    for i in range(ndim):
        if isinstance(idx[i], int):
            if idx[i] >= shape[i]:
                raise IndexError(
                    f'Index {idx[i]} is out of bounds'
                    f' for axis {i} with size {shape[i]}')
            new_idx.append(slice(idx[i], idx[i] + 1, 1))
        elif isinstance(idx[i], slice):
            start, stop, step = idx[i].indices(shape[i])
            if step <= 0:
                raise ValueError('Slice step must be positive.')
            if start == stop:
                raise ValueError(f'The index is empty on axis {i}')
            new_idx.append(slice(start, stop, step))
        else:
            raise ValueError(f'Invalid index on axis {i}')

    return tuple(new_idx)


def _normalize_index_map(
    shape: tuple[int, ...], index_map: dict[int, Any],
) -> dict[int, list[tuple[slice, ...]]]:
    new_index_map: dict[int, list[tuple[slice, ...]]] = {}
    for dev, idxs in index_map.items():
        if not isinstance(idxs, list):
            idxs = [idxs]

        idxs = [_normalize_index(shape, idx) for idx in idxs]
        idxs.sort()
        new_index_map[dev] = idxs

    return new_index_map



from libc.stdint cimport intptr_t, uintmax_t


###############################################################################
# Classes
###############################################################################

cdef class PointerAttributes:
    cdef:
        readonly int device
        readonly intptr_t devicePointer
        readonly intptr_t hostPointer
        readonly int type

cdef class MemPoolProps:
    # flatten the struct & list meaningful members
    cdef:
        int allocType
        int handleType
        int locationType
        int devId


###############################################################################
# Types and Enums
###############################################################################

IF CUPY_USE_CUDA_PYTHON:
    from cuda.ccudart cimport *
    # Aliases for compatibillity with existing CuPy codebase.
    # Keep in sync with names defined in `_runtime_typedef.pxi`.
    # TODO(kmaehashi): Remove these aliases.
    ctypedef cudaError_t Error
    ctypedef cudaDataType_t DataType

    ctypedef cudaDeviceAttr DeviceAttr
    ctypedef cudaMemoryAdvise MemoryAdvise

    ctypedef cudaStream_t Stream
    ctypedef cudaStreamCallback_t StreamCallback
    ctypedef cudaHostFn_t HostFn

    ctypedef cudaChannelFormatKind ChannelFormatKind
    ctypedef cudaChannelFormatDesc ChannelFormatDesc
    ctypedef cudaTextureObject_t TextureObject
    ctypedef cudaSurfaceObject_t SurfaceObject
    ctypedef cudaResourceType ResourceType
    ctypedef cudaTextureAddressMode TextureAddressMode
    ctypedef cudaTextureFilterMode TextureFilterMode
    ctypedef cudaTextureReadMode TextureReadMode
    ctypedef cudaResourceViewDesc ResourceViewDesc
    ctypedef cudaArray_t Array
    ctypedef cudaExtent Extent
    ctypedef cudaPos Pos
    ctypedef cudaPitchedPtr PitchedPtr
    ctypedef cudaMemcpyKind MemoryKind
    ctypedef cudaMipmappedArray_t MipmappedArray

    ctypedef cudaLimit Limit

    ctypedef cudaResourceDesc ResourceDesc

    ctypedef cudaMemcpy3DParms Memcpy3DParms

    ctypedef cudaTextureDesc TextureDesc

    ctypedef cudaIpcMemHandle_t IpcMemHandle

    ctypedef cudaIpcEventHandle_t IpcEventHandle

    ctypedef cudaUUID_t cudaUUID

    ctypedef cudaMemPool_t MemPool

    ctypedef cudaMemPoolAttr MemPoolAttr

    ctypedef cudaMemPoolProps _MemPoolProps

    ctypedef cudaPointerAttributes _PointerAttributes

    ctypedef cudaDeviceProp DeviceProp

    ctypedef cudaStreamCaptureStatus StreamCaptureStatus
    ctypedef cudaStreamCaptureMode StreamCaptureMode
    ctypedef cudaGraph_t Graph
    ctypedef cudaGraphExec_t GraphExec
    ctypedef cudaGraphNode_t GraphNode

    ctypedef cudaMemAllocationType MemAllocationType
    ctypedef cudaMemAllocationHandleType MemAllocationHandleType
    ctypedef cudaMemLocationType MemLocationType

ELSE:
    include "_runtime_typedef.pxi"
    from cupy_backends.cuda.api._runtime_enum cimport *


# For backward compatibility, keep APIs not prefixed with "cuda".
cpdef enum:
    memcpyHostToHost = 0
    memcpyHostToDevice = 1
    memcpyDeviceToHost = 2
    memcpyDeviceToDevice = 3
    memcpyDefault = 4

    hostAllocDefault = 0
    hostAllocPortable = 1
    hostAllocMapped = 2
    hostAllocWriteCombined = 4

    # cudaStream flags
    streamDefault = 0
    streamNonBlocking = 1

    # cudaStream handles
    streamLegacy = 1
    streamPerThread = 2

    eventDefault = 0
    eventBlockingSync = 1
    eventDisableTiming = 2
    eventInterprocess = 4

    # cudaStreamCaptureMode
    streamCaptureModeGlobal = 0
    streamCaptureModeThreadLocal = 1
    streamCaptureModeRelaxed = 2

    # cudaStreamCaptureStatus
    streamCaptureStatusNone = 0
    streamCaptureStatusActive = 1
    streamCaptureStatusInvalidated = 2

    # cudaMemoryType
    memoryTypeUnregistered = 0
    memoryTypeHost = 1
    memoryTypeDevice = 2
    memoryTypeManaged = 3


###############################################################################
# Constants
###############################################################################

# TODO(kmaehashi): Deprecate these aliases and use `cuda*`.
cdef int errorMemoryAllocation
cdef int errorInvalidValue
cdef int errorPeerAccessAlreadyEnabled
cdef int errorContextIsDestroyed
cdef int errorInvalidResourceHandle

cdef int deviceAttributeComputeCapabilityMajor
cdef int deviceAttributeComputeCapabilityMinor


###############################################################################
# Constants (CuPy)
###############################################################################

cdef bint _is_hip_environment


###############################################################################
# Error handling
###############################################################################

cpdef check_status(int status)


###############################################################################
# Initialization
###############################################################################

cpdef int driverGetVersion() except? -1
cpdef int runtimeGetVersion() except? -1
cpdef int _getCUDAMajorVersion() except? -1


###############################################################################
# Device and context operations
###############################################################################

cpdef int getDevice() except? -1
cpdef int deviceGetAttribute(int attrib, int device) except? -1
cpdef int deviceGetByPCIBusId(str pci_bus_id) except? -1
cpdef str deviceGetPCIBusId(int device)
cpdef int getDeviceCount() except? -1
cpdef setDevice(int device)
cpdef deviceSynchronize()
cpdef getDeviceProperties(int device)

cpdef int deviceCanAccessPeer(int device, int peerDevice) except? -1
cpdef deviceEnablePeerAccess(int peerDevice)
cpdef _deviceEnsurePeerAccess(int peerDevice)

cpdef size_t deviceGetLimit(int limit) except? -1
cpdef deviceSetLimit(int limit, size_t value)


###############################################################################
# Memory management
###############################################################################

cpdef intptr_t malloc(size_t size) except? 0
cpdef intptr_t mallocManaged(size_t size, unsigned int flags=*) except? 0
cpdef intptr_t malloc3DArray(intptr_t desc, size_t width, size_t height,
                             size_t depth, unsigned int flags=*) except? 0
cpdef intptr_t mallocArray(intptr_t desc, size_t width, size_t height,
                           unsigned int flags=*) except? 0
cpdef intptr_t mallocAsync(size_t size, intptr_t stream) except? 0
cpdef intptr_t mallocFromPoolAsync(size_t, intptr_t, intptr_t) except? 0
cpdef intptr_t hostAlloc(size_t size, unsigned int flags) except? 0
cpdef hostRegister(intptr_t ptr, size_t size, unsigned int flags)
cpdef hostUnregister(intptr_t ptr)
cpdef free(intptr_t ptr)
cpdef freeHost(intptr_t ptr)
cpdef freeArray(intptr_t ptr)
cpdef freeAsync(intptr_t ptr, intptr_t stream)
cpdef memGetInfo()
cpdef memcpy(intptr_t dst, intptr_t src, size_t size, int kind)
cpdef memcpyAsync(intptr_t dst, intptr_t src, size_t size, int kind,
                  intptr_t stream)
cpdef memcpyPeer(intptr_t dst, int dstDevice, intptr_t src, int srcDevice,
                 size_t size)
cpdef memcpyPeerAsync(intptr_t dst, int dstDevice,
                      intptr_t src, int srcDevice,
                      size_t size, intptr_t stream)
cpdef memcpy2D(intptr_t dst, size_t dpitch, intptr_t src, size_t spitch,
               size_t width, size_t height, MemoryKind kind)
cpdef memcpy2DAsync(intptr_t dst, size_t dpitch, intptr_t src, size_t spitch,
                    size_t width, size_t height, MemoryKind kind,
                    intptr_t stream)
cpdef memcpy2DFromArray(intptr_t dst, size_t dpitch, intptr_t src,
                        size_t wOffset, size_t hOffset, size_t width,
                        size_t height, int kind)
cpdef memcpy2DFromArrayAsync(intptr_t dst, size_t dpitch, intptr_t src,
                             size_t wOffset, size_t hOffset, size_t width,
                             size_t height, int kind, intptr_t stream)
cpdef memcpy2DToArray(intptr_t dst, size_t wOffset, size_t hOffset,
                      intptr_t src, size_t spitch, size_t width, size_t height,
                      int kind)
cpdef memcpy2DToArrayAsync(intptr_t dst, size_t wOffset, size_t hOffset,
                           intptr_t src, size_t spitch, size_t width,
                           size_t height, int kind, intptr_t stream)
cpdef memcpy3D(intptr_t Memcpy3DParmsPtr)
cpdef memcpy3DAsync(intptr_t Memcpy3DParmsPtr, intptr_t stream)
cpdef memset(intptr_t ptr, int value, size_t size)
cpdef memsetAsync(intptr_t ptr, int value, size_t size, intptr_t stream)
cpdef memPrefetchAsync(intptr_t devPtr, size_t count, int dstDevice,
                       intptr_t stream)
cpdef memAdvise(intptr_t devPtr, size_t count, int advice, int device)
cpdef PointerAttributes pointerGetAttributes(intptr_t ptr)
cpdef intptr_t deviceGetDefaultMemPool(int) except? 0
cpdef intptr_t deviceGetMemPool(int) except? 0
cpdef deviceSetMemPool(int, intptr_t)
cpdef intptr_t memPoolCreate(MemPoolProps) except? 0
cpdef memPoolDestroy(intptr_t)
cpdef memPoolTrimTo(intptr_t, size_t)
cpdef memPoolGetAttribute(intptr_t, int)
cpdef memPoolSetAttribute(intptr_t, int, object)


###############################################################################
# Stream and Event
###############################################################################

cpdef intptr_t streamCreate() except? 0
cpdef intptr_t streamCreateWithFlags(unsigned int flags) except? 0
cpdef streamDestroy(intptr_t stream)
cpdef streamSynchronize(intptr_t stream)
cpdef streamAddCallback(intptr_t stream, callback, intptr_t arg,
                        unsigned int flags=*)
cpdef launchHostFunc(intptr_t stream, callback, intptr_t arg)
cpdef _launchHostFuncUnmanaged(intptr_t stream, tuple func_arg)
cpdef streamQuery(intptr_t stream)
cpdef streamWaitEvent(intptr_t stream, intptr_t event, unsigned int flags=*)
cpdef streamBeginCapture(intptr_t stream, int mode=*)
cpdef intptr_t streamEndCapture(intptr_t stream) except? 0
cpdef bint streamIsCapturing(intptr_t stream) except*
cpdef intptr_t eventCreate() except? 0
cpdef intptr_t eventCreateWithFlags(unsigned int flags) except? 0
cpdef eventDestroy(intptr_t event)
cpdef float eventElapsedTime(intptr_t start, intptr_t end) except? 0
cpdef eventQuery(intptr_t event)
cpdef eventRecord(intptr_t event, intptr_t stream)
cpdef eventSynchronize(intptr_t event)


##############################################################################
# util
##############################################################################

cdef _ensure_context()


##############################################################################
# Texture
##############################################################################

cpdef uintmax_t createTextureObject(
    intptr_t ResDesc, intptr_t TexDesc) except? 0
cpdef destroyTextureObject(uintmax_t texObject)
cdef ChannelFormatDesc getChannelDesc(intptr_t array) except*
cdef ResourceDesc getTextureObjectResourceDesc(uintmax_t texobj) except*
cdef TextureDesc getTextureObjectTextureDesc(uintmax_t texobj) except*
cdef Extent make_Extent(size_t w, size_t h, size_t d) except*
cdef Pos make_Pos(size_t x, size_t y, size_t z) except*
cdef PitchedPtr make_PitchedPtr(
    intptr_t d, size_t p, size_t xsz, size_t ysz) except*

cpdef uintmax_t createSurfaceObject(intptr_t ResDesc) except? 0
cpdef destroySurfaceObject(uintmax_t surfObject)
# TODO(leofang): add cudaGetSurfaceObjectResourceDesc


##############################################################################
# Graph
##############################################################################

cpdef graphDestroy(intptr_t graph)
cpdef graphExecDestroy(intptr_t graphExec)
cpdef intptr_t graphInstantiate(intptr_t graph) except? 0
cpdef graphLaunch(intptr_t graphExec, intptr_t stream)
cpdef graphUpload(intptr_t graphExec, intptr_t stream)


##############################################################################
# Profiler
##############################################################################

cpdef profilerStart()
cpdef profilerStop()



cpdef enum:
    # CUjitInputType
    CU_JIT_INPUT_CUBIN = 0
    CU_JIT_INPUT_PTX = 1
    CU_JIT_INPUT_FATBINARY = 2
    CU_JIT_INPUT_OBJECT = 3
    CU_JIT_INPUT_LIBRARY = 4
    CU_JIT_INPUT_NVVM = 5

    # CUjit_option
    CU_JIT_MAX_REGISTERS = 0
    CU_JIT_THREADS_PER_BLOCK = 1
    CU_JIT_WALL_TIME = 2
    CU_JIT_INFO_LOG_BUFFER = 3
    CU_JIT_INFO_LOG_BUFFER_SIZE_BYTES = 4
    CU_JIT_ERROR_LOG_BUFFER = 5
    CU_JIT_ERROR_LOG_BUFFER_SIZE_BYTES = 6
    CU_JIT_OPTIMIZATION_LEVEL = 7
    CU_JIT_TARGET_FROM_CUCONTEXT = 8
    CU_JIT_TARGET = 9
    CU_JIT_FALLBACK_STRATEGY = 10
    CU_JIT_GENERATE_DEBUG_INFO = 11
    CU_JIT_LOG_VERBOSE = 12
    CU_JIT_GENERATE_LINE_INFO = 13
    CU_JIT_CACHE_MODE = 14
    CU_JIT_NEW_SM3X_OPT = 15
    CU_JIT_FAST_COMPILE = 16
    CU_JIT_GLOBAL_SYMBOL_NAMES = 17
    CU_JIT_GLOBAL_SYMBOL_ADDRESSES = 18
    CU_JIT_GLOBAL_SYMBOL_COUNT = 19
    CU_JIT_LTO = 20
    CU_JIT_FTZ = 21
    CU_JIT_PREC_DIV = 22
    CU_JIT_PREC_SQRT = 23
    CU_JIT_FMA = 24
    CU_JIT_NUM_OPTIONS = 25

    # CUfunction_attribute
    CU_FUNC_ATTRIBUTE_MAX_THREADS_PER_BLOCK = 0
    CU_FUNC_ATTRIBUTE_SHARED_SIZE_BYTES = 1
    CU_FUNC_ATTRIBUTE_CONST_SIZE_BYTES = 2
    CU_FUNC_ATTRIBUTE_LOCAL_SIZE_BYTES = 3
    CU_FUNC_ATTRIBUTE_NUM_REGS = 4
    CU_FUNC_ATTRIBUTE_PTX_VERSION = 5
    CU_FUNC_ATTRIBUTE_BINARY_VERSION = 6
    CU_FUNC_ATTRIBUTE_CACHE_MODE_CA = 7
    CU_FUNC_ATTRIBUTE_MAX_DYNAMIC_SHARED_SIZE_BYTES = 8
    CU_FUNC_ATTRIBUTE_PREFERRED_SHARED_MEMORY_CARVEOUT = 9

    # CUresult
    CUDA_ERROR_INVALID_VALUE = 1

    # CUarray_format
    CU_AD_FORMAT_UNSIGNED_INT8 = 0x01
    CU_AD_FORMAT_UNSIGNED_INT16 = 0x02
    CU_AD_FORMAT_UNSIGNED_INT32 = 0x03
    CU_AD_FORMAT_SIGNED_INT8 = 0x08
    CU_AD_FORMAT_SIGNED_INT16 = 0x09
    CU_AD_FORMAT_SIGNED_INT32 = 0x0a
    CU_AD_FORMAT_HALF = 0x10
    CU_AD_FORMAT_FLOAT = 0x20

    # CUaddress_mode
    CU_TR_ADDRESS_MODE_WRAP = 0
    CU_TR_ADDRESS_MODE_CLAMP = 1
    CU_TR_ADDRESS_MODE_MIRROR = 2
    CU_TR_ADDRESS_MODE_BORDER = 3

    # CUfilter_mode
    CU_TR_FILTER_MODE_POINT = 0
    CU_TR_FILTER_MODE_LINEAR = 1

    # Constants
    CU_TRSA_OVERRIDE_FORMAT = 0x01

    CU_TRSF_READ_AS_INTEGER = 0x01
    CU_TRSF_NORMALIZED_COORDINATES = 0x02
    CU_TRSF_SRGB = 0x10




import cupy
from cupy import testing

import cupyx.scipy.signal  # NOQA
import cupyx.scipy.signal.windows  # NOQA

try:
    import scipy.signal  # NOQA
    import scipy.signal.windows  # NOQA
except ImportError:
    scipy = None

import numpy as np

import pytest
from math import gcd

padtype_options = ["mean", "median", "minimum", "maximum", "line"]

_upfirdn_modes = [
    'constant', 'wrap', 'edge', 'smooth', 'symmetric', 'reflect',
    'antisymmetric', 'antireflect', 'line',
]

padtype_options += _upfirdn_modes


@testing.with_requires('scipy')
class TestResample:
    @testing.numpy_cupy_allclose(scipy_name='scp')
    def test_basic(self, xp, scp):
        # Some basic tests

        # Regression test for issue #scipy/3603.
        # window.shape must equal to sig.shape[0]
        sig = xp.arange(128)
        num = 256
        win = scp.signal.get_window(('kaiser', 8.0), 160)

        with pytest.raises(ValueError):
            scp.signal.resample(sig, num, window=win)

        # Other degenerate conditions
        with pytest.raises(ValueError):
            scp.signal.resample_poly(sig, 'yo', 1)

        with pytest.raises(ValueError):
            scp.signal.resample_poly(sig, 1, 0)

        # with pytest.raises(ValueError):
        #     scp.signal.resample_poly(sig, 2, 1, padtype='')

        # with pytest.raises(ValueError):
        #     scp.signal.resample_poly(sig, 2, 1, padtype='mean', cval=10)

        # test for issue #scipy/6505 - should not modify window.shape
        # when axis ≠ 0
        sig2 = xp.tile(xp.arange(160), (2, 1))
        return scp.signal.resample(sig2, num, axis=-1, window=win).copy()

    @pytest.mark.parametrize('window', (None, 'hamming'))
    @pytest.mark.parametrize('N', (20, 19))
    @pytest.mark.parametrize('num', (100, 101, 10, 11))
    @testing.numpy_cupy_allclose(scipy_name='scp', atol=1e-7)
    def test_rfft(self, N, num, window, xp, scp):
        # Make sure the speed up using rfft gives the same result as the normal
        # way using fft
        results = []
        x = xp.linspace(0, 10, N, endpoint=False)
        y = xp.cos(-x ** 2 / 6.0)

        results.append(scp.signal.resample(y, num, window=window).copy())
        results.append(scp.signal.resample(
            y + 0j, num, window=window).real.copy())

        y = xp.array([xp.cos(-x**2 / 6.0), xp.sin(-x**2 / 6.0)])
        y_complex = y + 0j
        results.append(scp.signal.resample(
            y, num, axis=1, window=window).copy())
        results.append(
            scp.signal.resample(
                y_complex, num, axis=1, window=window).real.copy())
        return results

    @testing.numpy_cupy_allclose(scipy_name='scp', atol=1e-7)
    def test_input_domain(self, xp, scp):
        # Test if both input domain modes produce the same results.
        tsig = xp.arange(256) + 0j
        fsig = xp.fft.fft(tsig)
        num = 256
        return (scp.signal.resample(fsig, num, domain='freq'),
                scp.signal.resample(tsig, num, domain='time'))

    @pytest.mark.parametrize('nx', (1, 2, 3, 5, 8))
    @pytest.mark.parametrize('ny', (1, 2, 3, 5, 8))
    @pytest.mark.parametrize('dtype', ('float', 'complex'))
    @testing.numpy_cupy_allclose(scipy_name='scp')
    def test_dc(self, nx, ny, dtype, xp, scp):
        x = xp.array([1] * nx, dtype)
        y = scp.signal.resample(x, ny)
        return y.copy()

    @pytest.mark.parametrize('padtype', padtype_options)
    @testing.numpy_cupy_allclose(scipy_name='scp')
    @pytest.mark.skip(reason='cval and mode is not supported on upfirdn')
    def test_mutable_window(self, padtype, xp, scp):
        # Test that a mutable window is not modified
        impulse = xp.zeros(3)
        window = xp.random.RandomState(0).randn(2)
        scp.signal.resample_poly(impulse, 5, 1, window=window, padtype=padtype)
        # assert_array_equal(window, window_orig)
        return window

    @pytest.mark.parametrize('padtype', padtype_options)
    @testing.numpy_cupy_allclose(scipy_name='scp', atol=1e-5, rtol=1e-5)
    @pytest.mark.skip(reason='cval and mode is not supported on upfirdn')
    def test_output_float32(self, padtype, xp, scp):
        # Test that float32 inputs yield a float32 output
        x = xp.arange(10, dtype=xp.float32)
        h = xp.array([1, 1, 1], dtype=xp.float32)
        y = scp.signal.resample_poly(x, 1, 2, window=h, padtype=padtype)
        return y

    @pytest.mark.parametrize('padtype', padtype_options)
    @pytest.mark.parametrize('dtype', [cupy.float32, cupy.float64])
    @testing.numpy_cupy_allclose(scipy_name='scp', atol=1e-5, rtol=1e-5)
    @pytest.mark.skip(reason='cval and mode is not supported on upfirdn')
    def test_output_match_dtype(self, padtype, dtype, xp, scp):
        # Test that the dtype of x is preserved per issue #14733
        x = xp.arange(10, dtype=dtype)
        y = scp.signal.resample_poly(x, 1, 2, padtype=padtype)
        return y

    @pytest.mark.parametrize(
        "method, ext, padtype",
        [("fft", False, None)]
        # upfirdn does not support cval and mode
        # + list(
        #     product(
        #         ["polyphase"], [False, True], padtype_options,
        #     )
        # ),
    )
    @testing.numpy_cupy_allclose(scipy_name='scp', atol=1e-7)
    def test_resample_methods(self, method, ext, padtype, xp, scp):
        # Test resampling of sinusoids and random noise (1-sec)
        rate = 100
        rates_to = [49, 50, 51, 99, 100, 101, 199, 200, 201]

        # Sinusoids, windowed to avoid edge artifacts
        t = xp.arange(rate) / float(rate)
        freqs = xp.array((1., 10., 40.))[:, xp.newaxis]
        x = xp.sin(2 * xp.pi * freqs * t) * scp.signal.windows.hann(rate)
        results = []

        for rate_to in rates_to:
            if method == 'fft':
                y_resamps = scp.signal.resample(x, rate_to, axis=-1)
            else:
                if ext and rate_to != rate:
                    # Match default window design
                    g = gcd(rate_to, rate)
                    up = rate_to // g
                    down = rate // g
                    max_rate = max(up, down)
                    f_c = 1. / max_rate
                    half_len = 10 * max_rate
                    window = scp.signal.firwin(
                        2 * half_len + 1, f_c, window=('kaiser', 5.0))
                    polyargs = {'window': window, 'padtype': padtype}
                else:
                    polyargs = {'padtype': padtype}

                y_resamps = scp.signal.resample_poly(
                    x, rate_to, rate, axis=-1, **polyargs)

            results.append(y_resamps.copy())

        # More tests of fft method (Master 0.18.1 fails these)
        if method == 'fft':
            x1 = xp.array([1.+0.j, 0.+0.j])
            y1_test = scp.signal.resample(x1, 4)
            x2 = xp.array([1., 0.5, 0., 0.5])
            y2_test = scp.signal.resample(x2, 2)  # downsampling a real array
            results.append(y1_test.copy())
            results.append(y2_test.copy())

        return results

    @testing.numpy_cupy_allclose(scipy_name='scp')
    def test_poly_vs_filtfilt(self, xp, scp):
        # Check that up=1.0 gives same answer as filtfilt + slicing
        # random_state = xp.random.RandomState(17)
        try_types = (int, xp.float32, xp.complex64, float, complex)
        size = 10000
        down_factors = [2, 11, 79]
        results = []

        for dtype in try_types:
            x = testing.shaped_random((size,), xp, dtype, scale=1.0)
            # x = random_state.randn(size).astype(dtype)
            if dtype in (xp.complex64, xp.complex128):
                x += 1j * testing.shaped_random(
                    (size,), xp, xp.float32, scale=1.0)

            # resample_poly assumes zeros outside of signl, whereas filtfilt
            # can only constant-pad. Make them equivalent:
            x[0] = 0
            x[-1] = 0

            for down in down_factors:
                h = scp.signal.firwin(31, 1. / down, window='hamming')

                # Need to pass convolved version of filter to resample_poly,
                # since filtfilt does forward and backward, but resample_poly
                # only goes forward
                hc = scp.signal.convolve(h, h[::-1])
                y = scp.signal.resample_poly(x, 1, down, window=hc)
                results.append(y)
        return results

    @testing.numpy_cupy_allclose(scipy_name='scp')
    def test_correlate1d(self, xp, scp):
        results = []
        for down in [2, 4]:
            for nx in range(1, 40, down):
                for nweights in (32, 33):
                    x = testing.shaped_random((nx,), xp, xp.float64, scale=1.0)
                    weights = testing.shaped_random(
                        (nweights,), xp, xp.float64, scale=1.0)
                    y_s = scp.signal.resample_poly(
                        x, up=1, down=down, window=weights)
                    results.append(y_s)
        return results


@testing.with_requires('scipy')
class TestDecimate:
    @pytest.mark.parametrize('mod', [(cupy, cupyx.scipy), (np, scipy)])
    def test_bad_args(self, mod):
        xp, scp = mod
        x = xp.arange(12)

        with pytest.raises(TypeError):
            scp.signal.decimate(x, q=0.5, n=1)

        with pytest.raises(TypeError):
            scp.signal.decimate(x, q=2, n=0.5)

    @testing.numpy_cupy_allclose(scipy_name='scp')
    def test_basic_IIR(self, xp, scp):
        x = xp.arange(12)
        y = scp.signal.decimate(
            x, 2, n=1, ftype='iir', zero_phase=False).round()
        return y

    @testing.numpy_cupy_allclose(scipy_name='scp')
    def test_basic_FIR(self, xp, scp):
        x = xp.arange(12)
        y = scp.signal.decimate(
            x, 2, n=1, ftype='fir', zero_phase=False).round()
        return y

    @testing.numpy_cupy_allclose(scipy_name='scp')
    def test_shape(self, xp, scp):
        # Regression test for ticket #1480.
        z = xp.zeros((30, 30))
        d0 = scp.signal.decimate(z, 2, axis=0, zero_phase=False)
        d1 = scp.signal.decimate(z, 2, axis=1, zero_phase=False)
        return d0, d1

    @pytest.mark.xfail(reason='Sometimes it fails depending on hardware')
    @testing.numpy_cupy_allclose(scipy_name='scp')
    def test_phaseshift_FIR(self, xp, scp):
        return self._test_phaseshift(xp, scp, method='fir', zero_phase=False)

    @pytest.mark.xfail(reason='Sometimes it fails depending on hardware')
    @testing.numpy_cupy_allclose(scipy_name='scp')
    def test_zero_phase_FIR(self, xp, scp):
        return self._test_phaseshift(xp, scp, method='fir', zero_phase=True)

    @testing.numpy_cupy_allclose(scipy_name='scp', atol=1e-4, rtol=1e-4)
    def test_phaseshift_IIR(self, xp, scp):
        return self._test_phaseshift(xp, scp, method='iir', zero_phase=False)

    @testing.numpy_cupy_allclose(scipy_name='scp', atol=1e-4, rtol=1e-4)
    def test_zero_phase_IIR(self, xp, scp):
        return self._test_phaseshift(xp, scp, method='iir', zero_phase=True)

    def _test_phaseshift(self, xp, scp, method, zero_phase):
        rate = 120
        rates_to = [15, 20, 30, 40]  # q = 8, 6, 4, 3

        t_tot = int(100)  # Need to let antialiasing filters settle
        t = xp.arange(rate * t_tot + 1) / float(rate)

        # Sinusoids at 0.8*nyquist, windowed to avoid edge artifacts
        freqs = xp.array(rates_to) * 0.8 / 2
        d = (xp.exp(1j * 2 * xp.pi * freqs[:, xp.newaxis] * t)
             * scp.signal.windows.tukey(t.size, 0.1))

        results = []
        for rate_to in rates_to:
            q = rate // rate_to
            t_to = xp.arange(rate_to*t_tot+1) / float(rate_to)
            d_tos = (xp.exp(1j * 2 * xp.pi * freqs[:, xp.newaxis] * t_to)
                     * scp.signal.windows.tukey(t_to.size, 0.1))

            # Set up downsampling filters, match v0.17 defaults
            if method == 'fir':
                n = 30
                system = scp.signal.dlti(
                    scp.signal.firwin(n + 1, 1. / q, window='hamming'), 1.)
            elif method == 'iir':
                n = 8
                wc = 0.8*xp.pi/q
                system = scp.signal.dlti(*scp.signal.cheby1(n, 0.05, wc/xp.pi))

            # Calculate expected phase response, as unit complex vector
            if zero_phase is False:
                _, h_resps = scp.signal.freqz(
                    system.num, system.den, freqs / rate * 2 * xp.pi)
                h_resps /= xp.abs(h_resps)
            else:
                h_resps = xp.ones_like(freqs)

            y_resamps = scp.signal.decimate(
                d.real, q, n, ftype=system, zero_phase=zero_phase)
            results.append(y_resamps)

            # Get phase from complex inner product, like CSD
            h_resamps = xp.sum(d_tos.conj() * y_resamps, axis=-1)
            h_resamps /= xp.abs(h_resamps)
        return results

    @testing.numpy_cupy_allclose(scipy_name='scp', atol=1e-7)
    def test_auto_n(self, xp, scp):
        # Test that our value of n is a reasonable choice (depends on
        # the downsampling factor)
        sfreq = 100.
        n = 1000
        t = xp.arange(n) / sfreq
        # will alias for decimations (>= 15)
        x = xp.sqrt(2. / n) * xp.sin(2 * xp.pi * (sfreq / 30.) * t)
        x_out = scp.signal.decimate(x, 30, ftype='fir')
        return x_out

    @testing.with_requires('scipy>=1.10')
    @testing.numpy_cupy_allclose(scipy_name='scp', atol=5e-5, rtol=5e-5)
    def test_long_float32(self, xp, scp):
        # regression: gh-15072.  With 32-bit float and either lfilter
        # or filtfilt, this is numerically unstable
        x = scp.signal.decimate(xp.ones(10_000, dtype=np.float32), 10)
        return x

    @testing.with_requires('scipy>=1.10')
    @testing.numpy_cupy_allclose(scipy_name='scp')
    def test_float16_upcast(self, xp, scp):
        # float16 must be upcast to float64
        x = scp.signal.decimate(xp.ones(100, dtype=xp.float16), 10)
        return x

    @testing.numpy_cupy_allclose(scipy_name='scp')
    @pytest.mark.skip(
        reason='zpk2tf is returning real outputs instead of complex ones')
    def test_complex_iir_dlti(self, xp, scp):
        # regression: gh-17845
        # centre frequency for filter [Hz]
        fcentre = 50
        # filter passband width [Hz]
        fwidth = 5
        # sample rate [Hz]
        fs = 1e3

        z, p, k = scp.signal.butter(
            2, 2 * xp.pi * fwidth / 2, output='zpk', fs=fs)
        z = z.astype(complex) * xp.exp(2j * xp.pi * fcentre / fs)
        p = p.astype(complex) * xp.exp(2j * xp.pi * fcentre / fs)
        system = scp.signal.dlti(z, p, k)

        t = xp.arange(200) / fs

        # input
        u = (xp.exp(2j * xp.pi * fcentre * t)
             + 0.5 * xp.exp(-2j * xp.pi * fcentre * t))

        ynzp = scp.signal.decimate(u, 2, ftype=system, zero_phase=False)
        yzp = scp.signal.decimate(u, 2, ftype=system, zero_phase=True)
        return ynzp, yzp

    @testing.numpy_cupy_allclose(scipy_name='scp')
    @pytest.mark.skip(reason='roots does not support non-symmetric inputs')
    def test_complex_fir_dlti(self, xp, scp):
        # centre frequency for filter [Hz]
        fcentre = 50
        # filter passband width [Hz]
        fwidth = 5
        # sample rate [Hz]
        fs = 1e3
        numtaps = 20

        # FIR filter about 0Hz
        bbase = scp.signal.firwin(numtaps, fwidth/2, fs=fs)

        # rotate these to desired frequency
        zbase = xp.roots(bbase)
        zrot = zbase * xp.exp(2j * xp.pi * fcentre/fs)
        # FIR filter about 50Hz, maintaining passband gain of 0dB
        bz = bbase[0] * xp.poly(zrot)

        system = scp.signal.dlti(bz, 1)

        t = xp.arange(200) / fs

        # input
        u = (xp.exp(2j * xp.pi * fcentre * t)
             + 0.5 * xp.exp(-2j * xp.pi * fcentre * t))

        ynzp = scp.signal.decimate(u, 2, ftype=system, zero_phase=False)
        yzp = scp.signal.decimate(u, 2, ftype=system, zero_phase=True)
        return ynzp, yzp



from cupy._core.core cimport _ndarray_base


cdef _ndarray_base _ndarray_all(_ndarray_base self, axis, out, keepdims)
cdef _ndarray_base _ndarray_any(_ndarray_base self, axis, out, keepdims)
cdef _ndarray_base _ndarray_greater(_ndarray_base self, other)
cdef _ndarray_base _ndarray_greater_equal(_ndarray_base self, other)
cdef _ndarray_base _ndarray_less(_ndarray_base self, other)
cdef _ndarray_base _ndarray_less_equal(_ndarray_base self, other)
cdef _ndarray_base _ndarray_equal(_ndarray_base self, other)
cdef _ndarray_base _ndarray_not_equal(_ndarray_base self, other)



from cupyx.distributed.array._array import DistributedArray  # NOQA
from cupyx.distributed.array._array import distributed_array  # NOQA
from cupyx.distributed.array._linalg import make_2d_index_map  # NOQA
from cupyx.distributed.array._linalg import matmul  # NOQA
from cupyx.distributed.array._modes import Mode  # NOQA
from cupyx.distributed.array._modes import REPLICA, MIN, MAX, SUM, PROD  # NOQA



import warnings

import numpy

import cupy
from cupy._core import _routines_math as _math
from cupy._core import _fusion_thread_local
from cupy._core import internal


def sum(a, axis=None, dtype=None, out=None, keepdims=False):
    """Returns the sum of an array along given axes.

    Args:
        a (cupy.ndarray): Array to take sum.
        axis (int or sequence of ints): Axes along which the sum is taken.
        dtype: Data type specifier.
        out (cupy.ndarray): Output array.
        keepdims (bool): If ``True``, the specified axes are remained as axes
            of length one.

    Returns:
        cupy.ndarray: The result array.

    .. seealso:: :func:`numpy.sum`

    """
    if _fusion_thread_local.is_fusing():
        if keepdims:
            raise NotImplementedError(
                'cupy.sum does not support `keepdims` in fusion yet.')
        if dtype is None:
            func = _math.sum_auto_dtype
        else:
            func = _math._sum_keep_dtype
        return _fusion_thread_local.call_reduction(
            func, a, axis=axis, dtype=dtype, out=out)

    # TODO(okuta): check type
    return a.sum(axis, dtype, out, keepdims)


def prod(a, axis=None, dtype=None, out=None, keepdims=False):
    """Returns the product of an array along given axes.

    Args:
        a (cupy.ndarray): Array to take product.
        axis (int or sequence of ints): Axes along which the product is taken.
        dtype: Data type specifier.
        out (cupy.ndarray): Output array.
        keepdims (bool): If ``True``, the specified axes are remained as axes
            of length one.

    Returns:
        cupy.ndarray: The result array.

    .. seealso:: :func:`numpy.prod`

    """
    if _fusion_thread_local.is_fusing():
        if keepdims:
            raise NotImplementedError(
                'cupy.prod does not support `keepdims` in fusion yet.')
        if dtype is None:
            func = _math._prod_auto_dtype
        else:
            func = _math._prod_keep_dtype
        return _fusion_thread_local.call_reduction(
            func, a, axis=axis, dtype=dtype, out=out)

    # TODO(okuta): check type
    return a.prod(axis, dtype, out, keepdims)


def nansum(a, axis=None, dtype=None, out=None, keepdims=False):
    """Returns the sum of an array along given axes treating Not a Numbers
    (NaNs) as zero.

    Args:
        a (cupy.ndarray): Array to take sum.
        axis (int or sequence of ints): Axes along which the sum is taken.
        dtype: Data type specifier.
        out (cupy.ndarray): Output array.
        keepdims (bool): If ``True``, the specified axes are remained as axes
            of length one.

    Returns:
        cupy.ndarray: The result array.

    .. seealso:: :func:`numpy.nansum`

    """
    if _fusion_thread_local.is_fusing():
        if keepdims:
            raise NotImplementedError(
                'cupy.nansum does not support `keepdims` in fusion yet.')
        if a.dtype.char in 'FD':
            func = _math._nansum_complex_dtype
        elif dtype is None:
            func = _math._nansum_auto_dtype
        else:
            func = _math._nansum_keep_dtype
        return _fusion_thread_local.call_reduction(
            func, a, axis=axis, dtype=dtype, out=out)

    # TODO(okuta): check type
    return _math._nansum(a, axis, dtype, out, keepdims)


def nanprod(a, axis=None, dtype=None, out=None, keepdims=False):
    """Returns the product of an array along given axes treating Not a Numbers
    (NaNs) as zero.

    Args:
        a (cupy.ndarray): Array to take product.
        axis (int or sequence of ints): Axes along which the product is taken.
        dtype: Data type specifier.
        out (cupy.ndarray): Output array.
        keepdims (bool): If ``True``, the specified axes are remained as axes
            of length one.

    Returns:
        cupy.ndarray: The result array.

    .. seealso:: :func:`numpy.nanprod`

    """
    if _fusion_thread_local.is_fusing():
        if keepdims:
            raise NotImplementedError(
                'cupy.nanprod does not support `keepdims` in fusion yet.')
        if dtype is None:
            func = _math._nanprod_auto_dtype
        else:
            func = _math._nanprod_keep_dtype
        return _fusion_thread_local.call_reduction(
            func, a, axis=axis, dtype=dtype, out=out)

    # TODO(okuta): check type
    return _math._nanprod(a, axis, dtype, out, keepdims)


def cumsum(a, axis=None, dtype=None, out=None):
    """Returns the cumulative sum of an array along a given axis.

    Args:
        a (cupy.ndarray): Input array.
        axis (int): Axis along which the cumulative sum is taken. If it is not
            specified, the input is flattened.
        dtype: Data type specifier.
        out (cupy.ndarray): Output array.

    Returns:
        cupy.ndarray: The result array.

    .. seealso:: :func:`numpy.cumsum`

    """
    return _math.scan_core(a, axis, _math.scan_op.SCAN_SUM, dtype, out)


def cumprod(a, axis=None, dtype=None, out=None):
    """Returns the cumulative product of an array along a given axis.

    Args:
        a (cupy.ndarray): Input array.
        axis (int): Axis along which the cumulative product is taken. If it is
            not specified, the input is flattened.
        dtype: Data type specifier.
        out (cupy.ndarray): Output array.

    Returns:
        cupy.ndarray: The result array.

    .. seealso:: :func:`numpy.cumprod`

    """
    return _math.scan_core(a, axis, _math.scan_op.SCAN_PROD, dtype, out)


def nancumsum(a, axis=None, dtype=None, out=None):
    """Returns the cumulative sum of an array along a given axis treating Not a
    Numbers (NaNs) as zero.

    Args:
        a (cupy.ndarray): Input array.
        axis (int): Axis along which the cumulative sum is taken. If it is not
            specified, the input is flattened.
        dtype: Data type specifier.
        out (cupy.ndarray): Output array.

    Returns:
        cupy.ndarray: The result array.

    .. seealso:: :func:`numpy.nancumsum`
    """
    a = _replace_nan(a, 0, out=out)
    return cumsum(a, axis=axis, dtype=dtype, out=out)


def nancumprod(a, axis=None, dtype=None, out=None):
    """Returns the cumulative product of an array along a given axis treating
    Not a Numbers (NaNs) as one.

    Args:
        a (cupy.ndarray): Input array.
        axis (int): Axis along which the cumulative product is taken. If it is
            not specified, the input is flattened.
        dtype: Data type specifier.
        out (cupy.ndarray): Output array.

    Returns:
        cupy.ndarray: The result array.

    .. seealso:: :func:`numpy.nancumprod`
    """
    a = _replace_nan(a, 1, out=out)
    return cumprod(a, axis=axis, dtype=dtype, out=out)


_replace_nan_kernel = cupy._core._kernel.ElementwiseKernel(
    'T a, T val', 'T out', 'if (a == a) {out = a;} else {out = val;}',
    'cupy_replace_nan')


def _replace_nan(a, val, out=None):
    if out is None or a.dtype != out.dtype:
        out = cupy.empty_like(a)
    _replace_nan_kernel(a, val, out)
    return out


def diff(a, n=1, axis=-1, prepend=None, append=None):
    """Calculate the n-th discrete difference along the given axis.

    Args:
        a (cupy.ndarray): Input array.
        n (int): The number of times values are differenced. If zero, the input
            is returned as-is.
        axis (int): The axis along which the difference is taken, default is
            the last axis.
        prepend (int, float, cupy.ndarray): Value to prepend to ``a``.
        append (int, float, cupy.ndarray): Value to append to ``a``.

    Returns:
        cupy.ndarray: The result array.

    .. seealso:: :func:`numpy.diff`
    """

    if n == 0:
        return a
    if n < 0:
        raise ValueError(
            "order must be non-negative but got " + repr(n))

    a = cupy.asanyarray(a)
    nd = a.ndim
    axis = internal._normalize_axis_index(axis, nd)

    combined = []

    if prepend is not None:
        prepend = cupy.asanyarray(prepend)
        if prepend.ndim == 0:
            shape = list(a.shape)
            shape[axis] = 1
            prepend = cupy.broadcast_to(prepend, tuple(shape))
        combined.append(prepend)

    combined.append(a)

    if append is not None:
        append = cupy.asanyarray(append)
        if append.ndim == 0:
            shape = list(a.shape)
            shape[axis] = 1
            append = cupy.broadcast_to(append, tuple(shape))
        combined.append(append)

    if len(combined) > 1:
        a = cupy.concatenate(combined, axis)

    slice1 = [slice(None)] * nd
    slice2 = [slice(None)] * nd
    slice1[axis] = slice(1, None)
    slice2[axis] = slice(None, -1)
    slice1 = tuple(slice1)
    slice2 = tuple(slice2)

    op = cupy.not_equal if a.dtype == numpy.bool_ else cupy.subtract
    for _ in range(n):
        a = op(a[slice1], a[slice2])

    return a


def gradient(f, *varargs, axis=None, edge_order=1):
    """Return the gradient of an N-dimensional array.

    The gradient is computed using second order accurate central differences
    in the interior points and either first or second order accurate one-sides
    (forward or backwards) differences at the boundaries.
    The returned gradient hence has the same shape as the input array.

    Args:
        f (cupy.ndarray): An N-dimensional array containing samples of a scalar
            function.
        varargs (list of scalar or array, optional): Spacing between f values.
            Default unitary spacing for all dimensions. Spacing can be
            specified using:

            1. single scalar to specify a sample distance for all dimensions.
            2. N scalars to specify a constant sample distance for each
               dimension. i.e. `dx`, `dy`, `dz`, ...
            3. N arrays to specify the coordinates of the values along each
               dimension of F. The length of the array must match the size of
               the corresponding dimension
            4. Any combination of N scalars/arrays with the meaning of 2. and
               3.

            If `axis` is given, the number of varargs must equal the number of
            axes. Default: 1.
        edge_order ({1, 2}, optional): The gradient is calculated using N-th
            order accurate differences at the boundaries. Default: 1.
        axis (None or int or tuple of ints, optional): The gradient is
            calculated only along the given axis or axes. The default
            (axis = None) is to calculate the gradient for all the axes of the
            input array. axis may be negative, in which case it counts from the
            last to the first axis.

    Returns:
        gradient (cupy.ndarray or list of cupy.ndarray): A set of ndarrays
        (or a single ndarray if there is only one dimension) corresponding
        to the derivatives of f with respect to each dimension. Each
        derivative has the same shape as f.

    .. seealso:: :func:`numpy.gradient`
    """
    f = cupy.asanyarray(f)
    ndim = f.ndim  # number of dimensions
    axes = internal._normalize_axis_indices(axis, ndim, sort_axes=False)

    len_axes = len(axes)
    n = len(varargs)
    if n == 0:
        # no spacing argument - use 1 in all axes
        dx = [1.0] * len_axes
    elif n == 1 and cupy.ndim(varargs[0]) == 0:
        # single scalar for all axes
        dx = varargs * len_axes
    elif n == len_axes:
        # scalar or 1d array for each axis
        dx = list(varargs)
        for i, distances in enumerate(dx):
            if cupy.ndim(distances) == 0:
                continue
            elif cupy.ndim(distances) != 1:
                raise ValueError("distances must be either scalars or 1d")
            if len(distances) != f.shape[axes[i]]:
                raise ValueError(
                    "when 1d, distances must match "
                    "the length of the corresponding dimension"
                )
            if numpy.issubdtype(distances.dtype, numpy.integer):
                # Convert numpy integer types to float64 to avoid modular
                # arithmetic in np.diff(distances).
                distances = distances.astype(numpy.float64)
            diffx = cupy.diff(distances)
            # if distances are constant reduce to the scalar case
            # since it brings a consistent speedup
            if (diffx == diffx[0]).all():  # synchronize
                diffx = diffx[0]
            dx[i] = diffx
    else:
        raise TypeError("invalid number of arguments")

    if edge_order > 2:
        raise ValueError("'edge_order' greater than 2 not supported")

    # use central differences on interior and one-sided differences on the
    # endpoints. This preserves second order-accuracy over the full domain.

    outvals = []

    # create slice objects --- initially all are [:, :, ..., :]
    slice1 = [slice(None)] * ndim
    slice2 = [slice(None)] * ndim
    slice3 = [slice(None)] * ndim
    slice4 = [slice(None)] * ndim

    otype = f.dtype
    if numpy.issubdtype(otype, numpy.inexact):
        pass
    else:
        # All other types convert to floating point.
        # First check if f is a numpy integer type; if so, convert f to float64
        # to avoid modular arithmetic when computing the changes in f.
        if numpy.issubdtype(otype, numpy.integer):
            f = f.astype(numpy.float64)
        otype = numpy.float64

    for axis, ax_dx in zip(axes, dx):
        if f.shape[axis] < edge_order + 1:
            raise ValueError(
                "Shape of array too small to calculate a numerical gradient, "
                "at least (edge_order + 1) elements are required."
            )
        # result allocation
        out = cupy.empty_like(f, dtype=otype)

        # spacing for the current axis
        uniform_spacing = cupy.ndim(ax_dx) == 0

        # Numerical differentiation: 2nd order interior
        slice1[axis] = slice(1, -1)
        slice2[axis] = slice(None, -2)
        slice3[axis] = slice(1, -1)
        slice4[axis] = slice(2, None)

        if uniform_spacing:
            out[tuple(slice1)] = (f[tuple(slice4)] - f[tuple(slice2)]) / (
                2.0 * ax_dx
            )
        else:
            dx1 = ax_dx[0:-1]
            dx2 = ax_dx[1:]
            dx_sum = dx1 + dx2
            a = -(dx2) / (dx1 * dx_sum)
            b = (dx2 - dx1) / (dx1 * dx2)
            c = dx1 / (dx2 * dx_sum)
            # fix the shape for broadcasting
            shape = [1] * ndim
            shape[axis] = -1
            a.shape = b.shape = c.shape = tuple(shape)
            # 1D equivalent -- out[1:-1] = a * f[:-2] + b * f[1:-1] + c * f[2:]
            out[tuple(slice1)] = (a * f[tuple(slice2)] +
                                  b * f[tuple(slice3)] +
                                  c * f[tuple(slice4)])

        # Numerical differentiation: 1st order edges
        if edge_order == 1:
            slice1[axis] = 0
            slice2[axis] = 1
            slice3[axis] = 0
            dx_0 = ax_dx if uniform_spacing else ax_dx[0]
            # 1D equivalent -- out[0] = (f[1] - f[0]) / (x[1] - x[0])
            out[tuple(slice1)] = (f[tuple(slice2)] - f[tuple(slice3)]) / dx_0

            slice1[axis] = -1
            slice2[axis] = -1
            slice3[axis] = -2
            dx_n = ax_dx if uniform_spacing else ax_dx[-1]
            # 1D equivalent -- out[-1] = (f[-1] - f[-2]) / (x[-1] - x[-2])
            out[tuple(slice1)] = (f[tuple(slice2)] - f[tuple(slice3)]) / dx_n

        # Numerical differentiation: 2nd order edges
        else:
            slice1[axis] = 0
            slice2[axis] = 0
            slice3[axis] = 1
            slice4[axis] = 2
            if uniform_spacing:
                a = -1.5 / ax_dx
                b = 2.0 / ax_dx
                c = -0.5 / ax_dx
            else:
                dx1 = ax_dx[0]
                dx2 = ax_dx[1]
                dx_sum = dx1 + dx2
                a = -(2.0 * dx1 + dx2) / (dx1 * (dx_sum))
                b = dx_sum / (dx1 * dx2)
                c = -dx1 / (dx2 * (dx_sum))
            # 1D equivalent -- out[0] = a * f[0] + b * f[1] + c * f[2]
            out[tuple(slice1)] = (a * f[tuple(slice2)] +
                                  b * f[tuple(slice3)] +
                                  c * f[tuple(slice4)])

            slice1[axis] = -1
            slice2[axis] = -3
            slice3[axis] = -2
            slice4[axis] = -1
            if uniform_spacing:
                a = 0.5 / ax_dx
                b = -2.0 / ax_dx
                c = 1.5 / ax_dx
            else:
                dx1 = ax_dx[-2]
                dx2 = ax_dx[-1]
                dx_sum = dx1 + dx2
                a = (dx2) / (dx1 * (dx_sum))
                b = -dx_sum / (dx1 * dx2)
                c = (2.0 * dx2 + dx1) / (dx2 * (dx_sum))
            # 1D equivalent -- out[-1] = a * f[-3] + b * f[-2] + c * f[-1]
            out[tuple(slice1)] = (a * f[tuple(slice2)] +
                                  b * f[tuple(slice3)] +
                                  c * f[tuple(slice4)])
        outvals.append(out)

        # reset the slice object in this dimension to ":"
        slice1[axis] = slice(None)
        slice2[axis] = slice(None)
        slice3[axis] = slice(None)
        slice4[axis] = slice(None)

    if len_axes == 1:
        return outvals[0]
    else:
        return outvals


def ediff1d(arr, to_end=None, to_begin=None):
    """
    Calculates the difference between consecutive elements of an array.

    Args:
        arr (cupy.ndarray): Input array.
        to_end (cupy.ndarray, optional): Numbers to append at the end
            of the returned differences.
        to_begin (cupy.ndarray, optional): Numbers to prepend at the
            beginning of the returned differences.

    Returns:
        cupy.ndarray: New array consisting differences among succeeding
        elements.

    .. seealso:: :func:`numpy.ediff1d`
    """
    if not isinstance(arr, cupy.ndarray):
        raise TypeError('`arr` should be of type cupy.ndarray')

    # to flattened array.
    arr = arr.ravel()

    # to ensure the dtype of the output array is same as that of input.
    dtype_req = arr.dtype

    # if none optional cases are given
    if to_begin is None and to_end is None:
        return arr[1:] - arr[:-1]

    if to_begin is None:
        l_begin = 0
    else:
        if not isinstance(to_begin, cupy.ndarray):
            raise TypeError('`to_begin` should be of type cupy.ndarray')
        if not cupy.can_cast(to_begin, dtype_req, casting="same_kind"):
            raise TypeError("dtype of `to_begin` must be compatible "
                            "with input `arr` under the `same_kind` rule.")

        to_begin = to_begin.ravel()
        l_begin = len(to_begin)

    if to_end is None:
        l_end = 0
    else:
        if not isinstance(to_end, cupy.ndarray):
            raise TypeError('`to_end` should be of type cupy.ndarray')
        if not cupy.can_cast(to_end, dtype_req, casting="same_kind"):
            raise TypeError("dtype of `to_end` must be compatible "
                            "with input `arr` under the `same_kind` rule.")

        to_end = to_end.ravel()
        l_end = len(to_end)

    # calculating using in place operation
    l_diff = max(len(arr) - 1, 0)
    result = cupy.empty(l_diff + l_begin + l_end, dtype=arr.dtype)
    # Cupy does not support subclassing a ndarray
    # result = arr.__array_wrap__(result)
    if l_begin > 0:
        result[:l_begin] = to_begin
    if l_end > 0:
        result[l_begin + l_diff:] = to_end
    cupy.subtract(arr[1:], arr[:-1], result[l_begin:l_begin + l_diff])
    return result


# TODO(okuta): Implement cross


def trapz(y, x=None, dx=1.0, axis=-1):
    """
    Integrate along the given axis using the composite trapezoidal rule.
    Integrate `y` (`x`) along the given axis.

    Args:
        y (cupy.ndarray): Input array to integrate.
        x (cupy.ndarray): Sample points over which to integrate. If None equal
            spacing `dx` is assumed.
        dx (float): Spacing between sample points, used if `x` is None, default
            is 1.
        axis (int): The axis along which the integral is taken, default is
            the last axis.

    Returns:
        cupy.ndarray: Definite integral as approximated by the trapezoidal
        rule.

    .. seealso:: :func:`numpy.trapz`
    """
    if not isinstance(y, cupy.ndarray):
        raise TypeError('`y` should be of type cupy.ndarray')

    if x is None:
        d = dx
    else:
        if not isinstance(x, cupy.ndarray):
            raise TypeError('`x` should be of type cupy.ndarray')
        if x.ndim == 1:
            d = diff(x)
            # reshape to correct shape
            shape = [1] * y.ndim
            shape[axis] = d.shape[0]
            d = d.reshape(shape)
        else:
            d = diff(x, axis=axis)
    nd = y.ndim
    slice1 = [slice(None)] * nd
    slice2 = [slice(None)] * nd
    slice1[axis] = slice(1, None)
    slice2[axis] = slice(None, -1)
    product = d * (y[tuple(slice1)] + y[tuple(slice2)]) / 2.0
    try:
        ret = product.sum(axis)
    except ValueError:
        ret = cupy.add.reduce(product, axis)
    return ret


def product(a, axis=None, dtype=None, out=None, keepdims=False):
    warnings.warn('Please use `prod` instead.', DeprecationWarning)
    return prod(a, axis, dtype, out, keepdims)


def cumproduct(a, axis=None, dtype=None, out=None):
    warnings.warn('Please use `cumprod` instead.', DeprecationWarning)
    return cumprod(a, axis, dtype, out)



from cupy._core.core cimport _ndarray_base


cdef _ndarray_sort(_ndarray_base self, int axis)
cdef _ndarray_base _ndarray_argsort(_ndarray_base self, axis)
cdef _ndarray_partition(_ndarray_base self, kth, int axis)
cdef _ndarray_base _ndarray_argpartition(self, kth, axis)



import cupy
import numpy
import pytest

from cupy import testing
from cupy.exceptions import AxisError
# TODO (grlee77): use fft instead of fftpack once min. supported scipy >= 1.4
import cupyx.scipy.fft  # NOQA
import cupyx.scipy.fftpack  # NOQA
import cupyx.scipy.ndimage  # NOQA

try:
    # scipy.fft only available since SciPy 1.4.0
    import scipy.fft  # NOQA
except ImportError:
    pass

try:
    # These modules will be present in all supported SciPy versions
    import scipy
    import scipy.fftpack  # NOQA
    import scipy.ndimage  # NOQA
    scipy_version = numpy.lib.NumpyVersion(scipy.__version__)
except ImportError:
    pass


@testing.parameterize(
    *(
        testing.product(
            {
                "shape": [(32, 16), (31, 15)],
                "dtype": [numpy.float32, numpy.float64, numpy.complex64,
                          numpy.complex128],
                "shift": [1, -3, (5, 5.3), (3, 5)],
            }
        )
        + testing.product(
            {
                "shape": [(5, 16, 7), ],
                "dtype": [numpy.float32, numpy.float64, numpy.complex64,
                          numpy.complex128],
                "shift": [3, (-1, 2.5, 1)],
            }
        )
        + testing.product(
            {
                "shape": [(15, ), ],
                "dtype": [numpy.float32, numpy.float64, numpy.complex64,
                          numpy.complex128],
                "shift": [8.5, (5,)],
            }
        )
    )
)
@testing.with_requires("scipy")
class TestFourierShift:

    def _test_real_nd(self, xp, scp, x, real_axis):
        a = scp.fft.rfft(x, axis=real_axis)
        # complex-valued FFTs on all other axes
        complex_axes = tuple([ax for ax in range(x.ndim) if ax != real_axis])
        if complex_axes:
            a = scp.fft.fftn(a, axes=complex_axes)

        a = scp.ndimage.fourier_shift(
            a, self.shift, n=x.shape[real_axis], axis=real_axis)

        if complex_axes:
            a = scp.fft.ifftn(a, axes=complex_axes)
        a = scp.fft.irfft(a, axis=real_axis)
        if not x.dtype.kind == 'c':
            a = a.real
        return xp.ascontiguousarray(a)

    @testing.with_requires("scipy>=1.4.0")
    @testing.numpy_cupy_allclose(atol=1e-5, rtol=1e-5, scipy_name="scp")
    def test_real_fft_axis0(self, xp, scp):
        x = testing.shaped_random(self.shape, xp, self.dtype)
        if x.dtype.kind == 'c':
            # skip: can't use rfft on complex-valued x
            return x
        return self._test_real_nd(xp, scp, x, 0)

    @testing.with_requires("scipy>=1.4.0")
    @testing.numpy_cupy_allclose(atol=1e-5, rtol=1e-5, scipy_name="scp")
    def test_real_fft_axis1(self, xp, scp):
        x = testing.shaped_random(self.shape, xp, self.dtype)
        if x.dtype.kind == 'c' or x.ndim < 2:
            # skip: can't use rfft along axis 1 on complex-valued x or 1d x
            return x
        return self._test_real_nd(xp, scp, x, 1)

    @testing.numpy_cupy_allclose(atol=1e-5, rtol=1e-5, scipy_name="scp")
    def test_complex_fft(self, xp, scp):
        x = testing.shaped_random(self.shape, xp, self.dtype)
        a = scp.fftpack.fftn(x)
        a = scp.ndimage.fourier_shift(a, self.shift)
        a = scp.fftpack.ifftn(a)
        if not x.dtype.kind == 'c':
            a = a.real
        return xp.ascontiguousarray(a)

    @testing.numpy_cupy_allclose(atol=1e-5, rtol=1e-5, scipy_name="scp")
    def test_complex_fft_with_output(self, xp, scp):
        x = testing.shaped_random(self.shape, xp, self.dtype)
        a = scp.fftpack.fftn(x)
        scp.ndimage.fourier_shift(a, self.shift, output=a)
        a = scp.fftpack.ifftn(a)
        if not x.dtype.kind == 'c':
            a = a.real
        return xp.ascontiguousarray(a)


@testing.parameterize(
    *(
        testing.product(
            {
                "shape": [(32, 16), (31, 15)],
                "dtype": [numpy.float32, numpy.float64, numpy.complex64,
                          numpy.complex128],
                "sigma": [1, (5, 5.3), (3, 5)],
            }
        )
        + testing.product(
            {
                "shape": [(5, 16, 7), ],
                "dtype": [numpy.float32, numpy.float64, numpy.complex64,
                          numpy.complex128],
                "sigma": [3, (1, 2.5, 3)],
            }
        )
        + testing.product(
            {
                "shape": [(15, ), ],
                "dtype": [numpy.float32, numpy.float64, numpy.complex64,
                          numpy.complex128],
                "sigma": [8.5, (5,)],
            }
        )
    )
)
@testing.with_requires("scipy")
class TestFourierGaussian:

    def _test_real_nd(self, xp, scp, x, real_axis):
        a = scp.fft.rfft(x, axis=real_axis)
        # complex-valued FFTs on all other axes
        complex_axes = tuple([ax for ax in range(x.ndim) if ax != real_axis])
        if complex_axes:
            a = scp.fft.fftn(a, axes=complex_axes)

        a = scp.ndimage.fourier_gaussian(
            a, self.sigma, n=x.shape[real_axis], axis=real_axis)

        if complex_axes:
            a = scp.fft.ifftn(a, axes=complex_axes)
        a = scp.fft.irfft(a, axis=real_axis)
        if not x.dtype.kind == 'c':
            a = a.real
        return xp.ascontiguousarray(a)

    @testing.with_requires("scipy>=1.4.0")
    @testing.numpy_cupy_allclose(atol=1e-5, rtol=1e-5, scipy_name="scp")
    def test_real_fft_axis0(self, xp, scp):
        x = testing.shaped_random(self.shape, xp, self.dtype)
        if x.dtype.kind == 'c':
            # skip: can't use rfft on complex-valued x
            return x
        return self._test_real_nd(xp, scp, x, 0)

    @testing.with_requires("scipy>=1.4.0")
    @testing.numpy_cupy_allclose(atol=1e-5, rtol=1e-5, scipy_name="scp")
    def test_real_fft_axis1(self, xp, scp):
        x = testing.shaped_random(self.shape, xp, self.dtype)
        if x.dtype.kind == 'c' or x.ndim < 2:
            # skip: can't use rfft along axis 1 on complex-valued x or 1d x
            return x
        return self._test_real_nd(xp, scp, x, 1)

    @testing.numpy_cupy_allclose(atol=1e-5, rtol=1e-5, scipy_name="scp")
    def test_complex_fft(self, xp, scp):
        x = testing.shaped_random(self.shape, xp, self.dtype)
        a = scp.fftpack.fftn(x)
        a = scp.ndimage.fourier_gaussian(a, self.sigma)
        a = scp.fftpack.ifftn(a)
        if not x.dtype.kind == 'c':
            a = a.real
        return xp.ascontiguousarray(a)

    @testing.numpy_cupy_allclose(atol=1e-5, rtol=1e-5, scipy_name="scp")
    def test_complex_fft_with_output(self, xp, scp):
        x = testing.shaped_random(self.shape, xp, self.dtype)
        a = scp.fftpack.fftn(x)
        scp.ndimage.fourier_gaussian(a, self.sigma, output=a)
        a = scp.fftpack.ifftn(a)
        if not x.dtype.kind == 'c':
            a = a.real
        return xp.ascontiguousarray(a)


@testing.parameterize(
    *(
        testing.product(
            {
                "shape": [(32, 16), (31, 15)],
                "dtype": [numpy.float32, numpy.float64, numpy.complex64,
                          numpy.complex128],
                "size": [1, (5, 5.3), (3, 5)],
            }
        )
        + testing.product(
            {
                "shape": [(5, 16, 7), ],
                "dtype": [numpy.float32, numpy.float64, numpy.complex64,
                          numpy.complex128],
                "size": [3, (1, 2.5, 3)],
            }
        )
        + testing.product(
            {
                "shape": [(15, ), ],
                "dtype": [numpy.float32, numpy.float64, numpy.complex64,
                          numpy.complex128],
                "size": [8.5, (5,)],
            }
        )
    )
)
@testing.with_requires("scipy")
class TestFourierUniform:

    def _test_real_nd(self, xp, scp, x, real_axis):
        a = scp.fft.rfft(x, axis=real_axis)
        # complex-valued FFTs on all other axes
        complex_axes = tuple([ax for ax in range(x.ndim) if ax != real_axis])
        if complex_axes:
            a = scp.fft.fftn(a, axes=complex_axes)

        a = scp.ndimage.fourier_uniform(
            a, self.size, n=x.shape[real_axis], axis=real_axis)

        if complex_axes:
            a = scp.fft.ifftn(a, axes=complex_axes)
        a = scp.fft.irfft(a, axis=real_axis)
        if not x.dtype.kind == 'c':
            a = a.real
        return xp.ascontiguousarray(a)

    @testing.with_requires("scipy>=1.4.0")
    @testing.numpy_cupy_allclose(atol=1e-5, rtol=1e-5, scipy_name="scp")
    def test_real_fft_axis0(self, xp, scp):
        x = testing.shaped_random(self.shape, xp, self.dtype)
        if x.dtype.kind == 'c':
            # skip: can't use rfft on complex-valued x
            return x
        return self._test_real_nd(xp, scp, x, 0)

    @testing.with_requires("scipy>=1.4.0")
    @testing.numpy_cupy_allclose(atol=1e-5, rtol=1e-5, scipy_name="scp")
    def test_real_fft_axis1(self, xp, scp):
        x = testing.shaped_random(self.shape, xp, self.dtype)
        if x.dtype.kind == 'c' or x.ndim < 2:
            # skip: can't use rfft along axis 1 on complex-valued x or 1d x
            return x
        return self._test_real_nd(xp, scp, x, 1)

    @testing.numpy_cupy_allclose(atol=1e-5, rtol=1e-5, scipy_name="scp")
    def test_complex_fft(self, xp, scp):
        x = testing.shaped_random(self.shape, xp, self.dtype)
        a = scp.fftpack.fftn(x)
        a = scp.ndimage.fourier_uniform(a, self.size)
        a = scp.fftpack.ifftn(a)
        if not x.dtype.kind == 'c':
            a = a.real
        return xp.ascontiguousarray(a)

    @testing.numpy_cupy_allclose(atol=1e-5, rtol=1e-5, scipy_name="scp")
    def test_complex_fft_with_output(self, xp, scp):
        x = testing.shaped_random(self.shape, xp, self.dtype)
        a = scp.fftpack.fftn(x)
        scp.ndimage.fourier_uniform(a, self.size, output=a)
        a = scp.fftpack.ifftn(a)
        if not x.dtype.kind == 'c':
            a = a.real
        return xp.ascontiguousarray(a)


@testing.parameterize(
    *(
        testing.product(
            {
                'shape': [(32, 16), (31, 15)],
                'size': [1, (5, 5), (3, 5)],
            }
        )
        + testing.product(
            {
                'shape': [(5, 16, 7)],
                'size': [3, (1, 2, 4)],
            }
        )
        + testing.product(
            {
                'shape': [(15, ), ],
                'size': [8, (5,)],
            }
        )
    )
)
@testing.with_requires('scipy')
class TestFourierEllipsoid():
    def _test_real_nd(self, xp, scp, x, real_axis):
        if x.ndim == 1 and scipy_version < '1.5.3':
            # 1D case gives an incorrect result in SciPy < 1.5.3
            pytest.skip('scipy version to old')

        a = scp.fft.rfft(x, axis=real_axis)
        # complex-valued FFTs on all other axes
        complex_axes = tuple([ax for ax in range(x.ndim) if ax != real_axis])
        if complex_axes:
            a = scp.fft.fftn(a, axes=complex_axes)

        a = scp.ndimage.fourier_ellipsoid(
            a, self.size, n=x.shape[real_axis], axis=real_axis
        )

        if complex_axes:
            a = scp.fft.ifftn(a, axes=complex_axes)
        a = scp.fft.irfft(a, axis=real_axis)
        if not x.dtype.kind == 'c':
            a = a.real
        return xp.ascontiguousarray(a)

    @testing.with_requires('scipy>=1.4.0')
    @testing.for_all_dtypes(no_complex=True)
    @testing.numpy_cupy_allclose(atol=1e-5, rtol=1e-5, scipy_name='scp')
    def test_real_fft_axis0(self, xp, scp, dtype):
        x = testing.shaped_random(self.shape, xp, dtype)
        return self._test_real_nd(xp, scp, x, 0)

    @testing.with_requires('scipy>=1.4.0')
    @testing.for_all_dtypes(no_complex=True)
    @testing.numpy_cupy_allclose(atol=1e-5, rtol=1e-5, scipy_name='scp')
    def test_real_fft_axis1(self, xp, scp, dtype):
        x = testing.shaped_random(self.shape, xp, dtype)
        if x.ndim < 2:
            # skip: there is no axis=1 on 1d arrays
            return x
        return self._test_real_nd(xp, scp, x, 1)

    @testing.for_all_dtypes()
    @testing.numpy_cupy_allclose(atol=1e-5, rtol=1e-5, scipy_name='scp')
    def test_complex_fft(self, xp, scp, dtype):
        x = testing.shaped_random(self.shape, xp, dtype)
        if x.ndim == 1 and scipy_version < '1.5.3':
            # 1D case gives an incorrect result in SciPy < 1.5.3
            pytest.skip('scipy version to old')
        a = scp.fftpack.fftn(x)
        a = scp.ndimage.fourier_ellipsoid(a, self.size)
        a = scp.fftpack.ifftn(a)
        if not x.dtype.kind == 'c':
            a = a.real
        return xp.ascontiguousarray(a)

    @testing.for_all_dtypes()
    @testing.numpy_cupy_allclose(atol=1e-5, rtol=1e-5, scipy_name='scp')
    def test_complex_fft_with_output(self, xp, scp, dtype):
        x = testing.shaped_random(self.shape, xp, dtype)
        if x.ndim == 1 and scipy_version < '1.5.3':
            # 1D case gives an incorrect result in SciPy < 1.5.3
            pytest.skip('scipy version to old')
        a = scp.fftpack.fftn(x)
        scp.ndimage.fourier_ellipsoid(a.copy(), self.size, output=a)
        a = scp.fftpack.ifftn(a)
        if not x.dtype.kind == 'c':
            a = a.real
        return xp.ascontiguousarray(a)


@testing.with_requires('scipy')
class TestFourierEllipsoidInvalid():

    # SciPy < 1.5 raises ValueError instead of AxisError
    @testing.with_requires('scipy>=1.5.0')
    def test_0d_input(self):
        for xp, scp in zip((numpy, cupy), (scipy, cupyx.scipy)):
            with pytest.raises(AxisError):
                scp.ndimage.fourier_ellipsoid(xp.asarray(5.0), size=2)
        return

    def test_4d_input(self):
        # SciPy should raise here too because >3d isn't implemented, but
        # as of 1.5.4, it does not.
        shape = (4, 6, 8, 10)
        for xp, scp in zip((cupy,), (cupyx.scipy,)):
            with pytest.raises(NotImplementedError):
                scp.ndimage.fourier_ellipsoid(xp.ones(shape), size=2)
        return

    # SciPy < 1.5 raises ValueError instead of AxisError
    @testing.with_requires('scipy>=1.5.0')
    def test_invalid_axis(self):
        # SciPy should raise here too because >3d isn't implemented, but
        # as of 1.5.4, it does not.
        shape = (6, 8)
        for xp, scp in zip((numpy, cupy), (scipy, cupyx.scipy)):
            with pytest.raises(AxisError):
                scp.ndimage.fourier_ellipsoid(xp.ones(shape), 2, axis=2)
            with pytest.raises(AxisError):
                scp.ndimage.fourier_ellipsoid(xp.ones(shape), 2, axis=-3)
        return

    def test_invalid_size(self):
        # test size length mismatch
        shape = (6, 8)
        for xp, scp in zip((numpy, cupy), (scipy, cupyx.scipy)):
            with pytest.raises(RuntimeError):
                scp.ndimage.fourier_ellipsoid(xp.ones(shape), size=(2, 3, 4))
            with pytest.raises(RuntimeError):
                scp.ndimage.fourier_ellipsoid(xp.ones(shape), size=(4,))
        return



import cupy


def _first(arr, axis):
    """Return arr[..., 0:1, ...] where 0:1 is in the `axis` position

    """

    return cupy.take_along_axis(arr, cupy.array(0, ndmin=arr.ndim), axis)


def _isconst(x):
    """Check if all values in x are the same.  nans are ignored.
    x must be a 1d array. The return value is a 1d array
    with length 1, so it can be used in cupy.apply_along_axis.

    """

    y = x[~cupy.isnan(x)]
    if y.size == 0:
        return cupy.array([True])
    else:
        return (y[0] == y).all(keepdims=True)


def zscore(a, axis=0, ddof=0, nan_policy='propagate'):
    """Compute the z-score.

    Compute the z-score of each value in the sample, relative to
    the sample mean and standard deviation.

    Parameters
    ----------
    a : array-like
        An array like object containing the sample data
    axis : int or None, optional
        Axis along which to operate. Default is 0. If None,
        compute over the whole arrsy `a`
    ddof : int, optional
        Degrees of freedom correction in the calculation of the
        standard deviation. Default is 0
    nan_policy : {'propagate', 'raise', 'omit'}, optional
        Defines how to handle when input contains nan. 'propagate'
        returns nan, 'raise' throws an error, 'omit' performs
        the calculations ignoring nan values. Default is
        'propagate'. Note that when the value is 'omit',
        nans in the input also propagate to the output,
        but they do not affect the z-scores computed
        for the non-nan values

    Returns
    -------
    zscore : array-like
        The z-scores, standardized by mean and standard deviation of
        input array `a`

    """

    return zmap(a, a, axis=axis, ddof=ddof, nan_policy=nan_policy)


def zmap(scores, compare, axis=0, ddof=0, nan_policy='propagate'):
    """Calculate the relative z-scores.

    Return an array of z-scores, i.e., scores that are standardized
    to zero mean and unit variance, where mean and variance are
    calculated from the comparison array.

    Parameters
    ----------
    scores : array-like
        The input for which z-scores are calculated
    compare : array-like
        The input from which the mean and standard deviation of
        the normalization are taken; assumed to have the same
        dimension as `scores`
    axis : int or None, optional
        Axis over which mean and variance of `compare` are calculated.
        Default is 0. If None, compute over the whole array `scores`
    ddof : int, optional
        Degrees of freedom correction in the calculation of the
        standard deviation. Default is 0
    nan_policy : {'propagate', 'raise', 'omit'}, optional
        Defines how to handle the occurrence of nans in `compare`.
        'propagate' returns nan, 'raise' raises an exception, 'omit'
        performs the calculations ignoring nan values. Default is
        'propagate'. Note that when the value is 'omit', nans in `scores`
        also propagate to the output, but they do not affect the z-scores
        computed for the non-nan values

    Returns
    -------
    zscore : array-like
        Z-scores, in the same shape as `scores`

    """

    policies = ['propagate', 'raise', 'omit']

    if nan_policy not in policies:
        raise ValueError("nan_policy must be one of {%s}" %
                         ', '.join("'%s'" % s for s in policies))

    a = compare

    if a.size == 0:
        return cupy.empty(a.shape)

    if nan_policy == 'raise':
        contains_nan = cupy.isnan(cupy.sum(a))

        if contains_nan:  # synchronize!
            raise ValueError("The input contains nan values")

    if nan_policy == 'omit':
        if axis is None:
            mn = cupy.nanmean(a.ravel())
            std = cupy.nanstd(a.ravel(), ddof=ddof)
            isconst = _isconst(a.ravel())
        else:
            mn = cupy.nanmean(a, axis=axis, keepdims=True)
            std = cupy.nanstd(a, axis=axis, keepdims=True, ddof=ddof)
            isconst = (_first(a, axis) == a).all(axis=axis, keepdims=True)
    else:
        mn = a.mean(axis=axis, keepdims=True)
        std = a.std(axis=axis, ddof=ddof, keepdims=True)
        if axis is None:
            isconst = (a.ravel()[0] == a).all()
        else:
            isconst = (_first(a, axis) == a).all(axis=axis, keepdims=True)

    # Set std deviations that are 0 to 1 to avoid division by 0.
    std[isconst] = 1.0
    z = (scores - mn) / std

    # Set the outputs associated with a constant input to nan.
    z[cupy.broadcast_to(isconst, z.shape)] = cupy.nan
    return z
